/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "nodelist.h"
#include "intlist.h"
#include "doublelist.h"
#include "anchors.h"
#include "kernel.h"

/**
 * A straightforward implementation of the Anchors Hierarchy.  See the paper:
 *
 @InProceedings{moore-anchors,
  author = {Andrew Moore},
  title = {The Anchors Hierarchy: Using the Triangle Inequality to Survive High-Dimensional Data},
  Booktitle = {Proceedings of the Twelfth Conference on Uncertainty in Artificial Intelligence},
  Year = {2000},
  Pages = {397-405},
  Publisher = {AAAI Press}
 }
 *
 * Or http://www.ri.cmu.edu/pub_files/pub3/moore_andrew_2000_1/moore_andrew_2000_1.pdf
 *
 * (Or Google "Anchors Hierarchy")
 *
 */

INLINE void resort_points(Node* a);

Node* new_anchor(int N) {
    Node* node = (Node*) malloc(sizeof(Node));
	AnchorExtra* ae;
	node->extra = malloc(sizeof(AnchorExtra));
	ae = node->extra;
    node->children = NULL;
    node->Nchildren = 0;
    if (N > 0) {
        node->indices = (int*) malloc(N * sizeof(int));
    } else {
        node->indices = NULL;
    }
    node->N = N;
    ae->anchor = -1;
    ae->radius = -1;
    if (N > 0) {
        ae->distances = (double*) malloc(N * sizeof(double));
    } else {
        ae->distances = NULL;
    }
    return node;
}

// returns the particle indexed by particle_index (in the anchor)
INLINE double* get_anchor_particle(double* X, Node* node, int particle_index, int D) {
    return X + node->indices[particle_index]*D;
}

// returns the centre of the anchor
INLINE double* get_anchor_centre(double* X, Node* node, int D) {
    AnchorExtra* ae = (AnchorExtra*) node->extra;
    return X + ae->anchor*D;
}

// update the radius of the anchor
INLINE void update_radius(Node* a) {
    AnchorExtra* ae = (AnchorExtra*) a->extra;
    ae->radius = ae->distances[0];
	ae->anchor = a->indices[a->N - 1];
}

// get the radius of the anchor
INLINE double anchor_radius(Node* a) {
    AnchorExtra* ae = (AnchorExtra*) a->extra;
    return ae->radius;
}

// free memory associated with the anchors hierarchy
INLINE void free_anchors(Node* node) {
    AnchorExtra* ae = (AnchorExtra*) node->extra;
    if (ae->distances) {
        free(ae->distances);
    }
    free(node->extra);
    node->extra = NULL;
}

// remove the points from anchor that are referenced by stolenlist
INLINE void remove_points(Node* node, intlist* stolenlist) {    
    int Nstolen = intlist_size(stolenlist);
	int start = 0;
    int dest = 0;
    int i;
	AnchorExtra* ae = (AnchorExtra*) node->extra;
	int Nkeep;
    if (Nstolen == 0) {
        return; // nothing to do
    }
    for (i = 0; i <= Nstolen; i++) {
        int end;
        int ntake;
        if (i < Nstolen) {
            // copy points up until the next stolen one.
            end = intlist_get(stolenlist, i);
        } else {
            end = node->N;
        }
        ntake = end - start;
        // remove particles by moving memory
        if (ntake > 0) {
            memmove(node->indices + dest, node->indices + start,
                    sizeof(int) * ntake);
            memmove(ae->distances + dest, ae->distances + start,
                    sizeof(double) * ntake);
            dest += ntake;
        }
        start = end + 1;
    }
    update_radius(node);
    intlist_remove_all(stolenlist);
    
    Nkeep = node->N - Nstolen;
    // update the number of particles in anchor
    node->N -= Nstolen;
    // relinquish unneeded memory
    node->indices = (int*) realloc(node->indices, Nkeep * sizeof(int));
    ae->distances = (double*) realloc(ae->distances, Nkeep * sizeof(double));
}

// go through all the other anchors, stealing points that are closer to the
// new anchor (newone) than to the old one.  add these points to an anchorlist.
// remove the points from the old anchor.
INLINE void steal_points(double* X, int D, Node** nodes,
		  int Ndone, Node* newone) {
    int a;
    intlist* indlist = intlist_new(); // list of indices of stolen points
    doublelist* distlist = doublelist_new(); // list of distances of stolen points (to the new anchor)
    intlist* stolenlist = intlist_new(); // list of stolen points
    int Nnew;
    double* new_anchor_centre = get_anchor_centre(X, newone, D);
	AnchorExtra* ae = (AnchorExtra*) newone->extra;
    // for each anchor...
    for (a = 0; a < Ndone; a++) {
        double cutoff_dist;
        int i;
        double* anchor_centre = get_anchor_centre(X, nodes[a], D);
        // cutoff distance is half the distance between the new anchor and the old anchor.
        cutoff_dist = kernel_distance(anchor_centre, new_anchor_centre, D) / 2.0;
        // for each point in the (other) anchor
        for (i = 0; i < nodes[a]->N; i++) {
            double dist;
            AnchorExtra* ae = (AnchorExtra*) nodes[a]->extra;
            double particle_distance_from_centre = ae->distances[i];
			double* particle;
            // any points that are closer than "cutoff_dist" to the old anchor cannot
            // be stolen by the new anchor.
            if (particle_distance_from_centre < cutoff_dist) {
                break;
            }
            // compute the distance between this point and the new anchor.
            particle = get_anchor_particle(X, nodes[a], i, D);
            dist = kernel_distance(particle, new_anchor_centre, D);
            // if it's less than the distance to the old anchor, then...
            if (dist < particle_distance_from_centre || particle == new_anchor_centre) {
                // steal it!
                intlist_append(stolenlist, i);
                intlist_append(indlist, nodes[a]->indices[i]);
                doublelist_append(distlist, dist);
            }
        }
        remove_points(nodes[a], stolenlist);
    }

    // copy the stolen points into this anchor.
    Nnew = intlist_size(indlist);
    if (newone->indices) {
        free(newone->indices);
    }
    newone->indices = (int*) malloc(Nnew * sizeof(int));
    if (ae->distances) {
        free(ae->distances);
    }
    ae->distances = (double*) malloc(Nnew * sizeof(double));
    intlist_copy(indlist, Nnew, newone->indices);
    doublelist_copy(distlist, Nnew, ae->distances);
    newone->N = Nnew;

    resort_points(newone);

    intlist_free(indlist);
    doublelist_free(distlist);
    intlist_free(stolenlist);
}

/*
 * Splits anchors in the 'splittable' list until 'Nanchors' anchors are
 * created, or until each anchor contains fewer than 'Nmin' points.
 * Returns the new anchors in the 'anchors' list.
 */
INLINE void split_anchor_helper(nodelist* splittable, nodelist* anchors,
			    double* X, int D, int Nmin, int Nanchors) {
    /* The "splittable" list contains anchors that contain more than the maximum
       number of particles; they can be split further.
       The "anchors" list contains anchors that have fewer than the maximum number
       of particles; they cannot be split further.
       Note that both lists contain anchor* pointers.
    */
    int i, a;
    Node* tosplit = NULL;
    Node* newone;
    double rad;

    while (1) {
        int n_splittable = nodelist_size(splittable);
        int n_anchors = nodelist_size(anchors);
        int n_all = n_splittable + n_anchors;
		AnchorExtra* ae;
		Node** allanchors;
		Node* an;
        // if we've created "Nanchors" anchors, or if all the anchors we've created have no more
        // than "Nmin" particles, then we're done.
        if (n_all == Nanchors || n_splittable == 0) {
            // copy all elements from "splittable" to "anchors" and return.
            nodelist_append_list(anchors, splittable);
            //anchorlist_remove_all(splittable);
            return;
        }

        // look through all the 'splittable' anchors to find the anchor with the largest radius.
        rad = -1.0;
        for (a = 0; a < n_splittable; a++) {
            Node* an = nodelist_get(splittable, a);
            double r;
            r = anchor_radius(an);
            if (r > rad) {
    	       tosplit = an;
    	       rad = r;
            }
        }

        newone = new_anchor(0);
        // make the furthest point of the 'to-be-split' node be the anchor point of the new node.
        ae = (AnchorExtra*) newone->extra;
        ae->anchor = tosplit->indices[0];
        
        allanchors = (Node**) malloc(n_all * sizeof(Node*));
    	nodelist_copy(splittable, n_splittable, allanchors);
    	nodelist_copy(anchors, n_anchors, allanchors + n_splittable);

    	steal_points(X, D, allanchors, n_all, newone);
    
        // check through all the 'splittable' anchors and move any that are
        // no longer splittable into the 'anchors' list.
        // (anchors can become non-splittable because enough of their points get stolen
        //  that they drop below the maximum number of points to be splittable).
        for (i = 0; i < n_splittable; i++) {
            an = nodelist_get(splittable, i);
            if (an->N <= Nmin) {
                nodelist_append(anchors, an);
                nodelist_remove(splittable, an);
                i--;
                n_splittable--;
            }
        }
    
        // add the 'newone' into either the 'splittable' or 'anchors' list.
        if (newone->N <= Nmin) {
            nodelist_append(anchors, newone);
        } else {
            nodelist_append(splittable, newone);
        }
		free(allanchors);
    }
}

// split anchors under main so that there are Nanchors created OR each anchor has less than Nmin
// particles in it
INLINE void split_anchor(Node* main, double* X, int D, int Nmin, int Nanchors,
		  Node** result, int* Nresult) {
    nodelist* splittable;
    nodelist* anchors;
    int Nrtn;
	int i, d;

    splittable = nodelist_new();
    anchors    = nodelist_new();

    nodelist_append(splittable, main);

    split_anchor_helper(splittable, anchors, X, D, Nmin, Nanchors);

    Nrtn = nodelist_size(anchors);

    nodelist_copy(anchors, Nrtn, result);
    
    for (i = 0; i < Nrtn; i++) {
        Node* n = result[i];
        d = n->N;
    }

    nodelist_free(splittable);
    nodelist_free(anchors);
    
    for (i = 0; i < Nrtn; i++) {
        Node* n = result[i];
        d = n->N;
    }

    *Nresult = Nrtn;
}

// resort the points belonging to anchor a so that a's points are in decreasing order wrt distance
// from the anchor
INLINE void resort_points(Node* a) {
    int i, N;
	int* order;
	AnchorExtra* ae = (AnchorExtra*) a->extra;
    N = a->N;
	order = (int*) malloc(N * sizeof(int));

    for (i = 0; i < N; i++) {
        order[i] = i;
    }
    if (N > 1) {
        compare_weights_vector = ae->distances;
        qsort(order, N, sizeof(int), compare_weights);
        compare_weights_vector = NULL;
        
        permute_ints(a->indices, order, N);
        permute_doubles(ae->distances, order, N);
    }
    
    free(order);

    update_radius(a);
}

/*
 * look at all pairs of anchors, and merge the two that will produce a parent
 * anchor of smallest radius (lowest upper bound, actually).
 * 
 * do this until only one anchor remains.
 */
INLINE void merge_anchors(double* X, int D, Node** anchors, int Nanchors) {
    while (1) {
        int i, j;
        double dist;
        int besti = -1, bestj = -1;
        Node *parent, *child1, *child2;
        int d;
        double* target;
        double best;
        int ibest = -1;
		AnchorExtra* ae;
		double* parent_anchor_centre;

        if (Nanchors == 1) {
            return;
        }

        // find the pair of anchors whose merging gives the lowest upper bound on the radius of the
        // parent anchor
        dist = INF;
        for (i = 0; i < Nanchors; i++) {
            for (j = i+1; j < Nanchors; j++) {
                // an upper bound on the radius of the parent anchor
                double di = (anchor_radius(anchors[i]) + anchor_radius(anchors[j]) +
                              kernel_distance(get_anchor_centre(X, anchors[i], D),
                                              get_anchor_centre(X, anchors[j], D), D));
                if (di < dist) {
                    dist = di;
                    besti = i;
                    bestj = j;
                }
            }
        }

        child1 = anchors[besti];
        child2 = anchors[bestj];
        parent = new_anchor(child1->N + child2->N);
        parent->N = child1->N + child2->N;
    
        memcpy(parent->indices, child1->indices, sizeof(int) * child1->N);
        memcpy(parent->indices + child1->N, child2->indices, sizeof(int) * child2->N);
        parent->children = (Node**) malloc(2 * sizeof(Node*));
        parent->children[0] = child1;
        parent->children[1] = child2;
        parent->Nchildren = 2;
    
        // originally found the point closest to the average of the two child anchors.
        // changed to centroid of parent!
        // ideal is to minimize radius, no?
        target = (double*) malloc(D * sizeof(double));
        for (d = 0; d < D; d++) {
            target[d] = 0;
        }
        
//        {
//            double* child1_anchor;
//            double* child2_anchor;
//            child1_anchor = get_anchor_centre(X, child1, D);
//            child2_anchor = get_anchor_centre(X, child2, D);
//            
//            for (d = 0; d < D; d++) {
//                target[d] = (child1_anchor[d] + child2_anchor[d]) / 2.0;
//            }
//        }
        
        for (i = 0; i < parent->N; i++) {
            for (d = 0; d < D; d++) {
                target[d] += get_anchor_particle(X, parent, i, D)[d];
            }
        }
        for (d = 0; d < D; d++) {
            target[d] /= (double) parent->N;
        }
        

        best = INF; // +infinity
        for (i = 0; i < parent->N; i++) {
            double di = kernel_distance(target, get_anchor_particle(X, parent, i, D), D);
            if (di < best) {
	           best = di;
	           ibest = parent->indices[i];
            }
        }
        ae = (AnchorExtra*) parent->extra;
        ae->anchor = ibest;

        parent_anchor_centre = get_anchor_centre(X, parent, D);
        // recompute all the distances and re-sort the points.
        for (i = 0; i < parent->N; i++) {
            double di;
            di = kernel_distance(parent_anchor_centre, get_anchor_particle(X, parent, i, D), D);
            ae->distances[i] = di;
        }

        resort_points(parent);

        // remove both children from the anchors array and add the parent.
        anchors[besti] = parent;
        if (Nanchors > 2) {
            anchors[bestj] = anchors[Nanchors-1];
        }
        Nanchors--;
		free(target);
    }
}

// recursively build the hierarchy middle-out (section 3.1)
Node* build_anchors_hierarchy_recursive(double* X, int D, Node* main, int leaf_size) {
    int Nanchors;
    int a;
	Node** anchors;
	Node* root;

    // create <= sqrt(main->N) anchors.
    Nanchors = (int) ceil(sqrt(main->N));

    // first, we split the "main" anchor, until either we get sqrt(main->N)
    // anchors, or all the anchors contain not more than MIN_PARTICLES_PER_BOX
    // particles.
    anchors = (Node**) malloc(Nanchors * sizeof(Node*));
    
    split_anchor(main, X, D, leaf_size, Nanchors, anchors, &Nanchors);
    
    // next, we recurse on each of the anchors we've created, and connect that
    // subtree into the tree we're currently building.
    for (a = 0; a < Nanchors; a++) {
        Node* root;
		if (anchors[a]->N <= leaf_size) {
	       continue;
        }
        root = build_anchors_hierarchy_recursive(X, D, anchors[a], leaf_size);
        // replace the 'root' node of the subtree by the leaf node of the current tree.
        anchors[a] = root;
    }

    // finally, we agglomerate, merging the anchors we've created.
    // note that the leaf anchors get overwritten in the merging process;
    // they become accessible only via the root node.
    merge_anchors(X, D, anchors, Nanchors);
    
	root = anchors[0];
	free(anchors);
    return root;
}

// build an anchors hierarchy for the data (X) in R^D, where there are N data points and the given
// leaf size and extra size for each anchor
INLINE Node* build_anchors_hierarchy(double* X, int D, int N, int leaf_size) {
    double dist;
    int ibest = -1;
    double* centroid = (double*) malloc(D * sizeof(double));
    int i,d;
    Node* root, *newroot;
	AnchorExtra* ae;
    root = new_anchor(N);
    
    for (i=0; i<N; i++) {
        root->indices[i] = i;
    }

    for (d=0; d<D; d++) {
        centroid[d] = 0;
    }

    // find the point nearest to the centroid and make that the first anchor.
    for (i = 0; i < N; i++) {
        double* particle = get_anchor_particle(X, root, i, D);
        for (d = 0; d < D; d++) {
            centroid[d] += particle[d];
        }
    }
    for (d = 0; d < D; d++) {
        centroid[d] /= (double) N;
    }

    root->N = N;

    dist = INF;
    for (i = 0; i < N; i++) {
        double di;
        di = kernel_distance(centroid, get_anchor_particle(X, root, i, D), D);
        if (di < dist) {
            dist = di;
            ibest = root->indices[i];
        }
    }
    
    ae = (AnchorExtra*) root->extra;
    ae->anchor = ibest;

    // recompute distances to the anchor.
    for (i=0; i<N; i++) {
        ae->distances[i] = kernel_distance(get_anchor_centre(X, root, D),
                                             get_anchor_particle(X, root, i, D), D);
    }

    // sort this anchor's points in order of decreasing distance.
    resort_points(root);
    newroot = build_anchors_hierarchy_recursive(X, D, root, leaf_size);
	free(centroid);
    return newroot;
}

// returns the minimum distance between two nodes
INLINE double anchors_node_node_mindist(Node* xnode, Node* ynode, tree_globals* globals) {
    double *x1, *x2;
    int D = globals->D;
    double dist;
    AnchorExtra* aex = (AnchorExtra*) xnode->extra;
    AnchorExtra* aey = (AnchorExtra*) ynode->extra;
    x1 = globals->X + aex->anchor * D;
    x2 = globals->Y + aey->anchor * D;
    dist = kernel_distance(x1, x2, D);
    dist -= anchor_radius(xnode);
    dist -= anchor_radius(ynode);
    dist = dmax(dist, 0.0);
    return dist;
}

// returns the maximum distance between two nodes
INLINE double anchors_node_node_maxdist(Node* xnode, Node* ynode, tree_globals* globals) {
    double *x1, *x2;
    int D = globals->D;
    double dist;
    AnchorExtra* aex = (AnchorExtra*) xnode->extra;
    AnchorExtra* aey = (AnchorExtra*) ynode->extra;
    x1 = globals->X + aex->anchor * D;
    x2 = globals->Y + aey->anchor * D;
    dist = kernel_distance(x1, x2, D);
    dist += anchor_radius(xnode) + anchor_radius(ynode);
    return dist;
}

// returns the minimum distance between a (source) node and a (target) particle
INLINE double anchors_node_particle_mindist(Node* xnode, double* yparticle, tree_globals* globals) {
    double *x2;
    int D = globals->D;
    double dist;
    AnchorExtra* aex = (AnchorExtra*) xnode->extra;
    x2 = globals->X + aex->anchor * D;
    dist = kernel_distance(yparticle, x2, D);
    dist -= anchor_radius(xnode);
    dist = dmax(dist, 0.0);
    return dist;
}

// returns the minimum distance between a (source) node and a (target) particle
INLINE double anchors_node_particle_maxdist(Node* xnode, double* yparticle, tree_globals* globals) {
    double *x2;
    int D = globals->D;
    double dist;
    AnchorExtra* aex = (AnchorExtra*) xnode->extra;
    x2 = globals->X + aex->anchor * D;
    dist = kernel_distance(yparticle, x2, D);
    dist += anchor_radius(xnode);
    return dist;
}
