/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "boxtree.h"
#include "wirth.h"
#include "definitions.h"

INLINE void free_boxtree(Node* node) {
    BoxExtra* be = (BoxExtra*) node->extra;
    if (be->lowvals) {
        free(be->lowvals);
    }
    if (be->highvals) {
        free(be->highvals);
    }
    free(node->extra);
    node->extra = NULL;
}

INLINE Node* boxtree_create_node(int N, int D) {
    Node* node = (Node*) malloc(sizeof(Node));
    BoxExtra* be;
    node->extra = malloc(sizeof(BoxExtra));
    be = (BoxExtra*) node->extra;
    node->children = NULL;
    node->Nchildren = 0;
    node->indices = (int*) malloc(N * sizeof(int));
    node->N = N;
    be->lowvals  = (double*) malloc(D * sizeof(double));
    be->highvals = (double*) malloc(D * sizeof(double));
    return node;
}

INLINE void boxtree_tighten_box_limits(Node* node, double* X, int D) {
    int d;
    int n;
    int* inds = node->indices;
    BoxExtra* be = (BoxExtra*) node->extra;
    for (d=0; d<D; d++) {
        double high,low;
        high = -INF; // -infinity
        low  = INF; // +infinity
        for (n=0; n<node->N; n++) {
            low = dmin(low, X[inds[n]*D + d]);
            high = dmax(high, X[inds[n]*D + d]);
        }
        be->lowvals[d] = low;
        be->highvals[d] = high;
    }
}

void boxtree_recursive(double *X, int N, int D, Node* node, int d,
    int arity, int leafsize) {
    int i;

    // base case: this is a leaf node.
    if (N <= leafsize) {
        return;
    }

    place_in_boxes(node, N, X, D, (d % D), arity);

    for (i=0; i<node->Nchildren; i++) {
        Node* child;
        child = node->children[i];
        boxtree_recursive(X, child->N, D, child, d+1, arity, leafsize);
    }
}

INLINE Node* create_boxtree(double *X, int N, int D, int leafsize, int arity) {
    int i;
    Node* root;
    root = boxtree_create_node(N, D);
    for (i=0; i<N; i++) {
        root->indices[i] = i;
    }
    boxtree_tighten_box_limits(root, X, D);
    boxtree_recursive(X, N, D, root, 0, arity, leafsize);
    return root;
}

INLINE void minrep_range(double* minx, double* maxx, double* miny, double* maxy,
    double* r1, double* r2, int D) {
    int d;
    for (d = 0; d < D; d++) {
        if (maxx[d] < miny[d]) {
            r1[d] = miny[d];
            r2[d] = maxx[d];
        } else if (maxy[d] < minx[d]) {
            r1[d] = minx[d];
            r2[d] = maxy[d];
        } else {
            // nodes overlap!
            r1[d] = 0;
            r2[d] = 0;
        }
    }
}

INLINE void maxrep_range(double* minx, double* maxx, double* miny, double* maxy,
    double* r1, double* r2, int D) {
    int d;
    for (d = 0; d < D; d++) {
        if (maxy[d] - minx[d] > maxx[d] - miny[d]) {
            r1[d] = maxy[d];
            r2[d] = minx[d];
        } else {
            r1[d] = maxx[d];
            r2[d] = miny[d];
        }             
    }
}

INLINE double boxtree_node_node_mindist(Node* xnode, Node* ynode, tree_globals* globals) {
    int D = globals->D;
    BoxExtra* bex = (BoxExtra*) xnode->extra;
    BoxExtra* bey = (BoxExtra*) ynode->extra;
    double *mn1=bex->lowvals, *mn2=bey->lowvals, *mx1=bex->highvals, *mx2=bey->highvals;
    double* r1 = (double*) malloc(D * sizeof(double));
    double* r2 = (double*) malloc(D * sizeof(double));
	double dist;
    minrep_range(mn1,mx1,mn2,mx2,r1,r2,D);
    dist = kernel_distance(r1, r2, D);
	free(r1);
	free(r2);
	return dist;
}

INLINE double boxtree_node_node_maxdist(Node* xnode, Node* ynode, tree_globals* globals) {
    int D = globals->D;
    double* r1 = (double*) malloc(D * sizeof(double));
    double* r2 = (double*) malloc(D * sizeof(double));
    BoxExtra* bex = (BoxExtra*) xnode->extra;
    BoxExtra* bey = (BoxExtra*) ynode->extra;
    double *mn1=bex->lowvals, *mn2=bey->lowvals, *mx1=bex->highvals, *mx2=bey->highvals;
	double dist;
    maxrep_range(mn1,mx1,mn2,mx2,r1,r2,D);
    dist = kernel_distance(r1, r2, D);
	free(r1);
	free(r2);
	return dist;
}

INLINE double boxtree_node_particle_mindist(Node* xnode, double* yparticle, tree_globals* globals) {
    int D = globals->D;
    double* r1 = (double*) malloc(D * sizeof(double));
    double* r2 = (double*) malloc(D * sizeof(double));
    BoxExtra* bex = (BoxExtra*) xnode->extra;
	double dist;
    minrep_range(bex->lowvals, bex->highvals, yparticle, yparticle, r1, r2, D);
    dist = kernel_distance(r1, r2, D);
	free(r1);
	free(r2);
	return dist;
}

INLINE double boxtree_node_particle_maxdist(Node* xnode, double* yparticle, tree_globals* globals) {
    int D = globals->D;
    double* r1 = (double*) malloc(D * sizeof(double));
    double* r2 = (double*) malloc(D * sizeof(double));
    BoxExtra* bex = (BoxExtra*) xnode->extra;
	double dist;
    maxrep_range(bex->lowvals, bex->highvals, yparticle, yparticle, r1, r2, D);
    dist = kernel_distance(r1, r2, D);
	free(r1);
	free(r2);
	return dist;
}

INLINE void place_in_boxes(Node* node, int N, double* X, int D, int d, int NB) {
    int b, i, nchildren;
    int* nodeinds = NULL;
    double low, range;
    Node* child = NULL;
	    BoxExtra* be = (BoxExtra*) node->extra;

    // (possibly) over-allocate...
    nchildren = 0;
    node->children = (Node**) malloc(NB * sizeof(Node*));
    nodeinds = node->indices;
    range = (be->highvals[d] - be->lowvals[d]);
    low = be->lowvals[d];

    // Use Wirth's algorithm to find the kth-smallest element;
    // for NB-ary trees, do this NB times.
    // Must make a copy of the array before calling the algorithm since
    // it modifies the elements in-place.  Pull out the D-th dimension
    // coordinates into the copy.
    {
        double *tempvals = NULL;
        double *parray = NULL;
        int NP;
        int stepsize;
        int k;
        double lower;
        double upper;
        int ntaken = 0;
        int leftoff;
    
        tempvals = malloc(N * sizeof(double));
        for (i=0; i<N; i++) {
            tempvals[i] = X[D * nodeinds[i] + d];
        }
    
        stepsize = (int) ceil((double) N / (double) NB);
        k = stepsize;
        parray = tempvals;
        NP = N;
        upper = -INF;
        leftoff = 0;

        for (b=0; b<NB; b++) {
            int n, nchild;
            // select points to add to this child...
            lower = upper;
          
            //printf("child %i of %i, lower %g.\n", b, NB, lower);
            //printf("leftoff = %i\n", leftoff);
          
            if (b < (NB-1)) {
                upper = kth_smallest(parray, NP, stepsize-1);
                parray += stepsize;
                NP -= stepsize;
                nchild = stepsize;
                //printf("(1) upper %g, nchild %i\n", upper, nchild);
            } else {
                upper = INF;
                nchild = N - ntaken;
                //printf("(2) upper %g, nchild %i\n", upper, nchild);
            }
            child = boxtree_create_node(nchild, D);
            n=0;
    
            // grab points strictly within the bounds...
            for (i=0; i<N; i++) {
                double val = X[D * nodeinds[i] + d];
                if (val < upper  &&  val > lower) {
                    child->indices[n++] = nodeinds[i];
                    if (n == nchild) {
                        break;
                    }
                }
            }
    
            //printf("target %i, grabbed %i inside, nlow %i, nup %i\n",
            //nchild, n, nlow, nup);
    
            if (n < nchild) {
                // grab points equal to the lower bound.
                // (note that we start from "leftoff".
                for (i=leftoff; i<N; i++) {
                    double val = X[D * nodeinds[i] + d];
                    if (val == lower) {
                        child->indices[n++] = nodeinds[i];
                        if (n == nchild) {
                            break;
                        }
                    }
                }
            }
    
            //printf("target %i, grabbed %i.\n", nchild, n);
    
            if (n < nchild) {
                // grab points equal to the upper bound.
                for (i=0; i<N; i++) {
                    double val = X[D * nodeinds[i] + d];
                    if (val == upper) {
                        child->indices[n++] = nodeinds[i];
                        if (n == nchild) {
                            // set leftoff...
                            leftoff = i+1;
                            break;
                        }
                    }
                }
            }

            ntaken += n;    
            if (n != nchild) {
                printf("Expected %i, got %i\n", nchild, n);
            }
            boxtree_tighten_box_limits(child, X, D);
    
            node->children[nchildren] = child;
            nchildren++;
        }

        free(tempvals);
        tempvals = NULL;

        if (nchildren != NB) {
            node->children = (Node**) realloc(node->children, nchildren * sizeof(Node*));
        }
        node->Nchildren = nchildren;
        return;
    }
}

//double* compare_placeinboxes_X;
//int     compare_placeinboxes_D;
//int compare_fn_placeinboxes(const void* v1, const void* v2) {
//  int ind1, ind2;
//  double d1, d2;
//  ind1 = *((int*)v1);
//  ind2 = *((int*)v2);
//  d1 = compare_placeinboxes_X[compare_placeinboxes_D * ind1];
//  d2 = compare_placeinboxes_X[compare_placeinboxes_D * ind2];
//  if (d1 < d2)
//    return -1;
//  else if (d1 > d2)
//    return 1;
//  else
//    return 0;
//}
//
///// note that this version is commented out.  it is slower but definitely works - the
///// version above may still have some bugs.
//
//void place_in_boxes(Node* node, int N, double* X, int D, int d, int NB) {
//  int b, i, nchildren;
//  int* nodeinds;
//  double low, range;
//  Node* child;
//  int* tempinds;
//  BoxExtra* be = (BoxExtra*) node->extra;
//  // over-allocate...
//  nchildren = 0;
//  node->children = (Node**) malloc(NB * sizeof(Node*));
//  nodeinds = node->indices;
//  range = (be->highvals[d] - be->lowvals[d]);
//  low = be->lowvals[d];
//  
//  // hack - sort the points, then split them into (nearly) even
//  // children.
//  tempinds = (int*) malloc(N * sizeof(int));
//  for (i=0; i<N; i++)
//    tempinds[i] = nodeinds[i];
//
//  compare_placeinboxes_X = X + d;
//  compare_placeinboxes_D = D;
//  qsort(tempinds, N, sizeof(int), compare_fn_placeinboxes);
//
//  for (b=0; b<NB; b++) {
//    int start, finish, nchild;
//    start  = (int)floor(N * (double)b / (double)NB);
//    // non-inclusive
//    finish = (int)floor(N * (double)(b+1) / (double)NB);
//    nchild = finish - start;
//    if (nchild == 0)
//      continue;
//    child = boxtree_create_node(nchild, D);
//    for (i=0; i<nchild; i++) {
//      child->indices[i] = tempinds[start + i];
//    }
//      
//    boxtree_tighten_box_limits(child, X, D);
//
//    node->children[nchildren] = child;
//    nchildren++;
//  }
//  free(tempinds);
//
//  node->children = (Node**) realloc(node->children, nchildren * sizeof(Node*));
//  node->Nchildren = nchildren;
//}
