/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "generic.h"

#ifndef BOXTREE_H
#define BOXTREE_H

#define BOXTREE  0
#define KDTREE   1

/**
A boxtree node (kdtree, etc).

The members are arranged in a way that optimizes word alignment.
*/
struct BoxExtra {
    /** The lower limit of the bounding hypercube. */
    double* lowvals;
    /** The upper limit of the bounding hypercube. */
    double* highvals;
};
typedef struct BoxExtra BoxExtra;
//typedef BoxNode BoxTree;

/**
 * Create a boxtree, reserving extra space at each node.
 *   \param X: array of data points.
 *   \param N: number of data points.
 *   \param D: dimensionality of data points.
 *   \param leafsize: maximum number of points at leaf nodes.
 *   \param arity: maximum number of children per node.
 *   \param extra: number of bytes of extra space to reserve, per node.
 *
 *  \return a pointer to the root node.
 *
 * Note that you probably want to use <tt>create_kdtree</tt> rather than
 * this function.  This version creates a naive type of boxtree that
 * just splits along each dimension in order.
 */
Node* create_boxtree(double *X, int N, int D, int leafsize, int arity);
                        
/**
 * Destroys a boxtree.
 * \param tree: the tree to free.
 * \param free_extra: if non-NULL, then for each node in the tree, this function will
 * be called with the node as argument.
 * Use this to release memory or other resources that
 * you have reserved.  Note that if you allocated extra memory when the tree was created
 * (using the \c create_boxtree_with_extra function)
 * you DON'T have to free that memory; it will be done automatically.
*/
void free_boxtree(Node* tree);

// used by kdtree (and others):

/**
   \internal
   
   Creates children nodes and splits the given node's points
   among the children.

   Used by kdtree.c (and others).
   
   \param node the node to split
   \param N the number of point at node \c node
   \param X the data points
   \param D the dimensionality of the data points
   \param d the dimenion along which to split
   \param NB the number of (balanced) children to create
   \param extra the number of extra bytes to allocate for each node.

   The points belonging to node \c node are split into \c NB even groups along
   dimension \c d.  \c NB child nodes are created and assigned to node \c node.
*/
INLINE void place_in_boxes(Node* node, int N, double* X, int D, int d, int NB);


/**
   \internal

   Creates a boxtree node.
   
   \param N: the number of points contained by this node.
   \param D: the dimensionality of the data points.
   \param extra: the number of extra bytes to allocate for this node.

   \return a newly-allocate boxtree node.
*/
Node* boxtree_create_node(int N, int D);

/**
   \internal

   Non-recursively finds the bounding hypercube of data points in this box.
   The bounding hypercube is contained in the \c lowvals and \c highvals arrays
   of this node.
 */
void boxtree_tighten_box_limits(Node* node, double* X, int D);

double boxtree_node_node_mindist(Node* xnode, Node* ynode, tree_globals* globals);
double boxtree_node_node_maxdist(Node* xnode, Node* ynode, tree_globals* globals);
double boxtree_node_particle_mindist(Node* xnode, double* yparticle, tree_globals* globals);
double boxtree_node_particle_maxdist(Node* xnode, double* yparticle, tree_globals* globals);


#endif
