/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "dualtree.h"

void do_sum(double* X, double* Y, double* w, double* P, int NP, char* ker_type, int NX,
			int NY, int D, double eps_rel, double eps_abs, char* structure, double* infls,
			int influence_mode, int weights_mode) {
	struct timeval tv1;
	struct timeval tv2;
	double ms;
	printf("%s sum:\n", structure);
    gettimeofday(&tv1, NULL);
    nbody_sum(X, Y, w, P, NP, ker_type, NX, NY, D, eps_rel, eps_abs, structure, infls, influence_mode, weights_mode);
    gettimeofday(&tv2, NULL);
    ms = ms_between(&tv1, &tv2);
    printf("%s sum took %f ms.\n", structure, ms);
}

void do_max(double* X, double* Y, double* w, double* P, int NP, char* ker_type, int NX,
			int NY, int D, char* structure, int* inds, double* infls, int influence_mode, int weights_mode) {
	struct timeval tv1;
	struct timeval tv2;
	double ms;
	printf("%s max:\n", structure);
    gettimeofday(&tv1, NULL);
    nbody_max(X, Y, w, P, NP, ker_type, NX, NY, D, 25, structure, inds, infls, influence_mode, weights_mode);
    gettimeofday(&tv2, NULL);
    ms = ms_between(&tv1, &tv2);
    printf("%s max took %f ms.\n", structure, ms);
}

void resetData(int* mginds, double* mginfls, int NY) {
    int i;
    for (i = 0; i < NY; i++) {
        mginds[i] = -1;
        mginfls[i] = -INF;
    }
}

int main(int argc, char *argv[]) {
    int NX = 2000;
    int NY = 1000;
    int D = 3;
    int np = 0;
	double params[4];
    char* ker_type = "gaussian";
    int influence_mode = POSITIVE_INFLUENCE;
    int weights_mode = POSITIVE_WEIGHTS;
    char* wm = "positive";
    double eps_rel = 1e-3;
    double eps_abs = 1e-3;
    int ok = 1;

    int i;
    double* X;
    double* Y;
    double* w;
    double *mginfls, *naiveinfls;
    int *mginds, *naiveinds;

    X = (double*) malloc(NX * D * sizeof(double));
    Y = (double*) malloc(NY * D * sizeof(double));
    w = (double*) malloc(NX * sizeof(double));
  
    setup_XYw(X, NX, Y, NY, D, w, 0, 1, 0, 1, wm);

    mginds = (int*) malloc(NY * sizeof(int));
    naiveinds = (int*) malloc(NY * sizeof(int));

    mginfls = (double*) malloc(NY * sizeof(double));
    naiveinfls = (double*) malloc(NY * sizeof(double));
    
    for (i = 0; i < NY; i++) {
        mginds[i] = -1;
        naiveinds[i] = -1;
        mginfls[i] = -INF;
        naiveinfls[i] = -INF;
    }
    

    printf("NX = %i.  NY = %i.  D = %i.\n", NX, NY, D);
    
    printf("\nSUM\n\n");
    if (strcmp(ker_type, "gaussian") == 0) {
        np = 1;
        params[0] = 0.01;
    } else if (strcmp(ker_type, "laplacian") == 0) {
        np = 1;
        params[0] = 0.1;
    } else if (strcmp(ker_type, "lpq") == 0) {
        np = 3;
        params[0] = 25; params[1] = 2; params[2] = 1;
    } else if (strcmp(ker_type, "sigmoid") == 0) {
        np = 4;
        params[0] = 2; params[1] = 1; params[2] = 1; params[3] = 4;
    } else if (strcmp(ker_type, "polynomial") == 0) {
        np = 4;
        params[0] = 2; params[1] = 2; params[2] = 1; params[3] = 1;
    } else if (strcmp(ker_type, "dog") == 0) {
        np = 1;
        params[0] = 3;
    } else if (strcmp(ker_type, "gaussian_neg") == 0) {
        np = 1;
        params[0] = 2;
    }
    
	do_sum(X, Y, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "naive", naiveinfls, influence_mode, weights_mode);
    
	do_sum(X, Y, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "anchors", mginfls, influence_mode, weights_mode);
    ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok;
    resetData(mginds, mginfls, NY);
    
	do_sum(X, Y, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "kdtree", mginfls, influence_mode, weights_mode); 
    ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok;
    resetData(mginds, mginfls, NY);
    
    if (D <= 3 && strcmp(ker_type, "gaussian") == 0) {
		do_sum(X, Y, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "fgt", mginfls, influence_mode, weights_mode);
        ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok;
        resetData(mginds, mginfls, NY);
    }
    
    printf("\nMAX\n\n");
    
	do_max(X, Y, w, params, np, ker_type, NX, NY, D, "naive", naiveinds, naiveinfls, influence_mode, weights_mode);

	do_max(X, Y, w, params, np, ker_type, NX, NY, D, "anchors", mginds, mginfls, influence_mode, weights_mode);
    ok = verification_check_max(mginds, naiveinds, mginfls, naiveinfls, NY) && ok;
    resetData(mginds, mginfls, NY);

    do_max(X, Y, w, params, np, ker_type, NX, NY, D, "kdtree", mginds, mginfls, influence_mode, weights_mode);
    ok = verification_check_max(mginds, naiveinds, mginfls, naiveinfls, NY) && ok;
    resetData(mginds, mginfls, NY);
    
    if (D == 1 && weights_mode == POSITIVE_WEIGHTS && strcmp(ker_type, "gaussian") == 0) {
		do_max(X, Y, w, params, np, ker_type, NX, NY, D, "dt", mginds, mginfls, influence_mode, weights_mode);
        ok = verification_check_max(mginds, naiveinds, mginfls, naiveinfls, NY) && ok;
        resetData(mginds, mginfls, NY);
    }
    
    if (ok) {
        printf("\nAll passed!\n");
    } else {
        printf("\nSome FAILED!\n");
    }

    free(mginds);
    free(mginfls);
    free(naiveinds);
    free(naiveinfls);

    free(X);
    free(Y);
    free(w);

    return 0;
}
