/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "gauss.h"
#include "utils.h"

// ******************************************************************
void gausst(double* xat, double* yat, double* zat, double* str, int
	    natoms, double* xtarg, double* ytarg, double* ztarg,
	    double* pot, int ntarg, double delta, double tol, int
	    nterms) {
     // note no workspace stuff here, all allocated in lenchk and in
     // gafexp as needed
    int i, j, ip, nside, nmax, kdis, nfmax, nlmax, /*nallbx,*/ ntbox;
    double /*toler,*/ boxdim, r=0, z;
    double qsum;

    // dynamically allocated in lenchk
    int* iloc, *itofst, *itradr, *ibox;
    double* center, ****locexp;    

    // threshold the tolerance
    if (tol < 1E-20) {
        tol = 1E-20;
    } else if (tol > 1E-1) {
        tol = 1E-1;
    }

    boxdim = sqrt(0.5*delta);
    if (natoms > ntarg) {
        nmax = natoms;
    } else {
        nmax = ntarg;
    }

    nside = (int) (1.0/boxdim)+1;

    qsum = 0.0;
    for (i=0;i<natoms;i++) { 
        qsum = qsum + dabs(str[i]); 
    } 
    if (qsum==0.0) {
        qsum = 1;
    }

    // determine the number of terms. Note d = 3 here!!!
    if (nterms==0) {

        /*
        // from the original FGT paper; incorrect.
        {
            toler = tol;
            r = 1.0/(nside*sqrt(2.0*delta));
            ip = 0;
            z = qsum*pow(1.09,3.0)*pow(r,6.0)/pow(1-r,3);
            while (z > toler) {
                ip ++;
                z = pow(r,3.0)*z/pow((double)ip,3.0/2);
            }
            nterms = ip;
            printf("Old estimate: ip=%i, z=%g\n", ip, z);
        }
        */

        // from "A New Error Estimate of the Fast Gauss Transform",
        // Baxter and Roussos,
        // SIAM J.Scient.Comput. 24-1, 257--2579, 2002
        // http://epubs.siam.org/sam-bin/getfile/SISC/articles/39692.pdf
        {
            // see the paper for nice typeset equations. i'm using equation (12),
            // the left hand side of which I think is wrong - I think it's actually the
            //     sum_{j=1}^N w_j exp(| x - x_j |^2 / delta); ie, G(x); ie, the error
            // estimate is for the absolute error in the sum-product estimate for each target point.
    
            // |FGT - real| <= Q/((1-r)^d) sum_{k=0}^{d-1} (d choose k) (1-r^p)^k (r^p / sqrt(p!))^(d-k)
    
    
            // I make another change: there are TWO sources of error in the FGT: in
            // truncating the series expansion, and in determining how many boxes are
            // in the neighbourhood.  In order to guarantee that the total error is less
            // than the tolerance, the SUM of these two error sources must satisfy the
            // bound.  It could be possible to do a parameter search to optimize.  Instead,
            // I assign half the allowable error to each error source.

            int k, d;
            double acc;
            double dchoosek[] = { 1.0, 3.0, 3.0 };
            double sqrtfacp = 1.0; // sqrtfacp is sqrt(p!)
            d = 3;
            r = 1.0 / (ceil(sqrt(2.0 / delta)) * sqrt(2.0 * delta));

            // ip is p, the number of terms.
            for (ip=1;; ip++) {
            	sqrtfacp *= sqrt((double)ip);
            	acc = 0.0;
            	for (k=0; k<d; k++) {
            	   acc += dchoosek[k] * pow(1 - pow(r, (double)ip), (double)k)
            	       * pow( pow(r, (double)ip) / sqrtfacp, (double)(d-k) );
            	}
	            z = qsum / pow(1-r, (double)d) * acc;
            	// the truncation error is allowed to cause half the allowable error...
            	if (z < (0.5 * tol)) {
            	   break;
                }
            }
            /* printf("Newest estimate: ip=%i, z=%g\n", ip, z);
    	    printf("r=%g, delta=%g\n", r, delta); */
            nterms = ip;
            //printf("p=%i, err=%g, bound=%g\n", ip, z, tol/2.0);
        }
    }


    // this is the number of boxes (per dimension) considered to be in
    // the neighbourhood of a target box.
    // the exponential falloff of the gaussian allows boxes outside a certain
    // radius to be ignored.  note, however, that this radius depends on the
    // sum of weights!
    /*
    kdis = sqrt(-2*log(tol));
    printf("Old kdis = %i\n", kdis);
    */
    {
        double t;
        // the neighbourhood truncation is allowed to cause half the allowable error.
        // This formula comes from the original FGT paper of Greengard and Strain.
        t = -log( (0.5 * tol) / qsum) / (2.0 * r * r);
        if (t < 0.0) {
            kdis = 0;
        } else {
            kdis = (int)ceil(sqrt(t));
        }
        //kdis = sqrt(-log(tol / qsum) / (2.0 * r * r));
        //printf("kdis = %i\n", kdis);
        //printf("tol=%g, qsum=%g.\n", tol, qsum);
        //printf("kdis=%i, err=%g, bound=%g\n", kdis, qsum * exp(-2*r*r*kdis*kdis),
        //tol/2.0);
    }
    nfmax = nterms + 2;
    nlmax = nterms + 2;
 
    // allocate iloc, locexp, ibox, center, itofst, itradr and set up
    // iloc. the int ntbox is assigned there also
    lenchk(nside, xtarg, ytarg, ztarg, ntarg, nterms, nmax, &iloc, &ibox, &itofst,
           &itradr, &locexp, &center, &ntbox);

    /*
    alloced_mem.iloc = iloc;
    alloced_mem.ibox = ibox;
    alloced_mem.itofst = itofst;
    alloced_mem.itradr = itradr;
    alloced_mem.locexp = locexp;
    alloced_mem.center = center;
    alloced_mem.nterms = nterms;
    alloced_mem.ntbox = ntbox;
    alloced_mem.valid = 1;
    */

    // zero the potentials
    for (j=0;j<ntarg;j++) {
        pot[j]=0;
    }

    // call gafexp--
    gafexp(xat, yat, zat, str, natoms, xtarg, ytarg, ztarg, ntarg,
	       delta, nside, nterms, tol, iloc, ntbox, nfmax, nlmax, pot,
	       kdis, ibox, itradr, itofst, center, locexp);
  
    gaeval(xtarg, ytarg, ztarg, pot, ntarg, locexp, nterms, ibox,
	       center, delta, iloc, nlmax, nside, itofst, itradr);
 

    /*
        alloced_mem.valid = 0;
    */

    // out with the trash
    freestuff(iloc, ibox, itofst, itradr, locexp, center, nterms, ntbox);
    return;
}

// ******************************************************************
void gdir(double xp, double yp, double zp, double* xat, double* yat,
	  double* zat, int natoms, double* str, double* dpot, double
	  delta) {
    double x, y, z;
    int j;

    for (j=0;j<natoms;j++) {
        x = xp - xat[j];
        y = yp - yat[j];
        z = zp - zat[j];
        *dpot = *dpot + exp(-(x*x + y*y + z*z)/delta)*str[j];
    }
}

// ******************************************************************
void gamval(double*** ffexp, int nterms, double* cent, double xpoint,
	    double ypoint, double zpoint, double* pot, double delta) {
    int i,j,k, i2;
    double facx, facy, facz, x, y, z, x2, y2, z2, dsq, potinc, hi, hj;
    double* hexpx, *hexpy, *hexpz;

    hexpx = (double*) malloc((nterms+1) * sizeof(double));
    hexpy = (double*) malloc((nterms+1) * sizeof(double));
    hexpz = (double*) malloc((nterms+1) * sizeof(double));

    dsq = 1.0/sqrt(delta);
    x = (xpoint - cent[0])*dsq;
    y = (ypoint - cent[1])*dsq;
    z = (zpoint - cent[2])*dsq;

    x2 = 2.0*x;
    y2 = 2.0*y;
    z2 = 2.0*z;

    facx = exp(-x*x);
    facy = exp(-y*y);
    facz = exp(-z*z);

    // inintialize
    hexpx[0] = facx;
    hexpy[0] = facy;
    hexpz[0] = facz;

    hexpx[1] = x2*facx;
    hexpy[1] = y2*facy;
    hexpz[1] = z2*facz;

    for (i=1;i<nterms;i++) {
        i2 = 2*i;
        hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
        hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
        hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
    }

    potinc = 0.0;
    for (i=0;i<(nterms+1);i++) {
        hi = hexpx[i];
        for (j=0;j<(nterms+1);j++) {
            hj = hexpy[j];
            for (k=0;k<(nterms+1);k++) {
                potinc = potinc + hi*hj*hexpz[k]*ffexp[i][j][k];
            }
        }
    }

    *pot = *pot + potinc;

    free(hexpx);
    free(hexpy);
    free(hexpz);

}

// ******************************************************************
void gamkxp(double* cent, double* xatoms, double* yatoms, double* zatoms,
	    double* str, int ninbox, double*** ffexp, int nterms,
	    double delta) {
    int i,j,k,l;
    double dsq,x,y,z,xp;
    double* zp, *yp;

    zp = (double*) malloc((nterms+1) * sizeof(double));
    yp = (double*) malloc((nterms+1) * sizeof(double));

    for (i=0;i<(nterms+1);i++) {
        for (j=0;j<(nterms+1);j++) {
            for (k=0;k<(nterms+1);k++) {
                ffexp[i][j][k] = 0.0;
            }
        }
    }

    dsq = 1.0/sqrt(delta);
    for (i=0;i<ninbox;i++) {
        x = (xatoms[i] - cent[0])*dsq;
        y = (yatoms[i] - cent[1])*dsq;
        z = (zatoms[i] - cent[2])*dsq;
      
        zp[0] = 1.0;
        yp[0] = 1.0;
        for (k=1;k<(nterms+1);k++) {
            zp[k] = zp[k-1]*z/k;
            yp[k] = yp[k-1]*y/k;
        }
        xp = str[i];
        for (j=0;j<(nterms+1);j++) {
            for (k=0;k<(nterms+1);k++) {
                for (l=0;l<(nterms+1);l++) {
                    ffexp[j][k][l] = ffexp[j][k][l]+xp*yp[k]*zp[l];
                }
            }
            xp = xp*x/(j+1);
        } 
    }
    free(zp);
    free(yp);
    return;
}

// ******************************************************************
void galval(double*** local, int nterms, double* cent, double xpoint,
	    double ypoint, double zpoint, double* pot, double delta) {

    int i, j, k, l;
    double parsum1, parsum2, parsum3, x, y, z, potinc, dsq;

    dsq = 1.0/(sqrt(delta));
    x = (xpoint - cent[0])*dsq;
    y = (ypoint - cent[1])*dsq;
    z = (zpoint - cent[2])*dsq;

    // using 3-d horner's rule
    potinc = 0.0;
    for (i=nterms;i>0;i--) {
        parsum1 = 0.0;
        for (j=nterms;j>0;j--) {
            parsum2 = 0.0;
            for (k=nterms;k>0;k--) {
                parsum2 = (parsum2 + local[i][j][k])*z;
            }
            parsum2 = parsum2 + local[i][j][0];
            parsum1 = (parsum1 + parsum2)*y;
        }
        parsum3 = 0.0;
        for (l=nterms;l>0;l--) {
            parsum3 = (parsum3 + local[i][0][l])*z;
        }
        parsum3 = parsum3 + local[i][0][0];
        potinc = (potinc + parsum3 + parsum1)*x;
    }

    parsum1 = 0.0;
    for (j=nterms;j>0;j--) {
        parsum2 = 0.0;
        for (k=nterms;k>0;k--) {
            parsum2 = (parsum2 + local[0][j][k])*z;
        }
        parsum2 = parsum2 + local[0][j][0];
        parsum1 = (parsum1 + parsum2)*y;
    }
    potinc = potinc + parsum1;

    parsum1 = 0.0;
    for (k=nterms;k>0;k--) {
        parsum1 = (parsum1 + local[0][0][k])*z;
    }
    parsum1 = parsum1 + local[0][0][0];
    potinc = potinc + parsum1;

    *pot = potinc;
    return;
}


// ******************************************************************
void galkxp(double* cent, double xp, double yp, double zp, double
	    charge, double*** b, int nterms, double delta) {

    int i,l,m,n,i2;
    double dsq, x, y, z, x2, y2, z2, facx, facy, facz, /*prod,*/ facs, hl, hm;

    double* hexpx, *hexpy, *hexpz /* ,*fac */;
 
    hexpx = (double*) malloc((nterms+1) * sizeof(double));
    hexpy = (double*) malloc((nterms+1) * sizeof(double));
    hexpz = (double*) malloc((nterms+1) * sizeof(double));

    dsq = 1.0/sqrt(delta);

    // MAYBE WRONG!! I reversed them- makes sense see paper
    x = (cent[0] - xp)*dsq;
    x2 = 2.0*x;
    y = (cent[1] - yp)*dsq;
    y2 = 2.0*y;
    z = (cent[2] - zp)*dsq;
    z2 = 2.0*z;

    facx = exp(-x*x)*charge;
    facy = exp(-y*y);
    facz = exp(-z*z);

    hexpx[0] = facx;
    hexpy[0] = facy;
    hexpz[0] = facz;

    hexpx[1] = x2*facx;
    hexpy[1] = y2*facy;
    hexpz[1] = z2*facz;

    for (i=1;i<nterms;i++) {
        i2 = 2*i;
        hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
        hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
        hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
    }
  
    facs = 1.0;
    for (l=0;l<nterms+1;l++) {
        hexpx[l] = hexpx[l]*facs;
        hexpy[l] = hexpy[l]*facs;
        hexpz[l] = hexpz[l]*facs;
        facs = -1*facs/(l+1.0);
    }
    for (l=0;l<nterms+1;l++) {
        hl = hexpx[l];
        for (m=0;m<nterms+1;m++) {
            hm = hexpy[m];
            for (n=0;n<nterms+1;n++) {
                b[l][m][n] = hl*hm*hexpz[n];
            }
        }
    }
    free(hexpx);
    free(hexpy);
    free(hexpz);
    return;
}

// ******************************************************************
void gaeval(double* xtarg, double* ytarg, double* ztarg, double* pot,
	    int ntarg, double**** locexp, int nterms, int* ibox,
	    double* center, double delta, int* iloc, int nlmax, int
	    nside, int* itofst, int* itradr) {

    int i, k, jt, nboxes, inoff, jadr, ninbox;
    double xp, yp, zp;

    nboxes = nside*nside*nside;
    for (i=0;i<nboxes;i++) {
        inoff = itofst[i];
        ninbox = itofst[i+1] - inoff;
        if (ninbox > nlmax) {
            jadr = iloc[i];
            for (k=0; k< ninbox; k++) {
                jt = itradr[inoff+k];
                xp = xtarg[jt];
                yp = ytarg[jt];
                zp = ztarg[jt];
                galval(locexp[jadr], nterms, &center[3*i], xtarg[jt],
                       ytarg[jt], ztarg[jt], &pot[jt], delta);
            }
        }
    }
    return;
}


// ******************************************************************
void gafexp(double* xat, double* yat, double* zat, double* str, int
	    natoms, double* xtarg, double* ytarg, double* ztarg, int
	    ntarg, double delta, int nside, int nterms, double tol,
	    int* iloc, int ntbox, int nfmax, int nlmax, double* pot,
	    int kdis, int* ibox, int* itradr, int* itofst, double*
	    center, double**** locexp)  {
    double* xatoms, *yatoms, *zatoms, *charg2;
    double cent[3];
    double*** ffexp, ***b;
    double xp, yp, zp;
    int i, j, k, l, nboxes, maxpts, ioff, iadr, inoff, nnbors, ninnbr, jt, iadloc, ninbox;
    int*ioffst, *nbors;
    double*** temp1, ***temp2;

    // temp1,temp2 not allocated in gshift to avoid overhead each time
    // it's called

    //  iloc, ibox, itradr, itofst, locexp, center allocated in
    //  lenchk. iloc is assigned there as well

    nboxes = nside*nside*nside;
    if (natoms>ntarg) {
        maxpts = natoms;
    } else {
        maxpts = ntarg;
    }

    // dynamic allocations?
    ioffst = (int*) malloc((nboxes+1) * sizeof(int));
    nbors = (int*) malloc(nboxes * sizeof(int));
    xatoms = (double*) malloc(natoms * sizeof(double)); 
    yatoms = (double*) malloc(natoms * sizeof(double)); 
    zatoms  = (double*) malloc(natoms * sizeof(double)); 
    charg2  = (double*) malloc(natoms * sizeof(double)); 

    // locexp is allocated in lenchk

    // allocating the ffexp
    ffexp = (double***) malloc((nterms+1) * sizeof(double**));
    for (i=0;i<(nterms+1);i++) {
        ffexp[i] = (double **) malloc((nterms+1) * sizeof(double*));
        for (j=0;j<(nterms+1);j++) {
            ffexp[i][j] = (double *) malloc((nterms+1) * sizeof(double));
        }
    }

    // allocating b, temp1, temp2
    b = (double***) malloc((nterms+1) * sizeof(double**));
    temp1 = (double***) malloc((nterms+1) * sizeof(double**));
    temp2 = (double***) malloc((nterms+1) * sizeof(double**));
    for (i=0;i<(nterms+1);i++) {
        b[i] = (double**) malloc((nterms+1) * sizeof(double*));
        temp1[i] = (double**) malloc((nterms+1) * sizeof(double*));
        temp2[i] = (double**) malloc((nterms+1) * sizeof(double*));
        for (j=0;j<(nterms+1);j++) {
            b[i][j] = (double*) malloc((nterms+1) * sizeof(double));
            temp1[i][j] = (double*) malloc((nterms+1) * sizeof(double));
            temp2[i][j] = (double*) malloc((nterms+1) * sizeof(double));
        }
    }

    // call assign:
    assign(nside, xat, yat, zat, str, xatoms, yatoms, zatoms, charg2,
           ioffst, xtarg, ytarg, ztarg, ntarg, ibox, natoms, center,
           itradr, itofst); 

    // zero the local expansions
    for (i=0;i<ntbox;i++) {
        for (j=0;j<(nterms+1);j++) {
            for (k=0;k<(nterms+1);k++) {
                for (l=0;l<(nterms+1);l++) {
                    locexp[i][j][k][l] = 0;
                }
            }
        }
    }

    for (i=0;i<nboxes;i++) {
        ioff = ioffst[i];
        ninbox = ioffst[i+1] - ioff;
        if ((ninbox<=nfmax)&&(ninbox>0)) {
            // No far-field expansion
            mknbor(i, nbors, &nnbors, nside, kdis);
            for (j=0;j<nnbors;j++) {
                iadr = nbors[j];
                inoff = itofst[iadr];
                ninnbr = itofst[iadr+1]-inoff;
                if (ninnbr <= nlmax) {
                    // Direct interaction
                    for (k=0;k<ninnbr;k++) {
                        jt = itradr[inoff+k];
                        xp = xtarg[jt];
                        yp = ytarg[jt];
                        zp = ztarg[jt];
                        gdir(xp, yp, zp, &xatoms[ioff], &yatoms[ioff],
                             &zatoms[ioff], ninbox, &charg2[ioff],
                             &pot[jt], delta);
                    }
                } else {
                    // Sources in neighbors shifted to taylor series
                    cent[0] = center[3*iadr];
                    cent[1] = center[3*iadr+1];
                    cent[2] = center[3*iadr+2];
                    iadloc = iloc[iadr];
                    for (k=0;k<ninbox;k++) {
                        galkxp(cent, xatoms[ioff+k], yatoms[ioff+k],
                    	       zatoms[ioff+k], charg2[ioff+k], b, nterms,
                    	       delta);
                        addexp(b, locexp[iadloc], nterms); 		    
                    }
                }
            }
        } else if ((ninbox>nfmax)&&(ninbox>0)) {
            // create far-field expansion
            cent[0] = center[3*i];
            cent[1] = center[3*i+1];
            cent[2] = center[3*i+2];
            gamkxp(cent, &xatoms[ioff], &yatoms[ioff], &zatoms[ioff], &charg2[ioff], ninbox, ffexp, nterms, delta);
            mknbor(i, nbors, &nnbors, nside, kdis); 
            for (j=0;j<nnbors;j++) {
                iadr = nbors[j];
                inoff = itofst[iadr];
                ninnbr = itofst[iadr+1] - inoff;
                if (ninnbr <= nlmax) { 
                    // too few targets, shift to local
                    for (k=0;k<ninnbr;k++) {
                        jt = itradr[inoff+k];
                        xp = xtarg[jt];
                        yp = ytarg[jt];
                        zp = ztarg[jt];
                        gamval(ffexp, nterms, cent, xp, yp, zp, &pot[jt], delta);
                    }
                } else  {
                    gshift(ffexp, cent, &center[3*iadr], b, nterms, temp1, temp2, delta); 
                    iadloc = iloc[iadr];
                    addexp(b, locexp[iadloc], nterms);
                }
    	    }
    	}
    }
    
 
    // free ffexp, b, temp1, temp2
    for(i = 0; i < (nterms+1); i++) {
        for (j=0; j < (nterms+1);j++) {
            free(ffexp[i][j]);
            free(b[i][j]);
            free(temp1[i][j]);
            free(temp2[i][j]);
        }
        free(ffexp[i]);
        free(b[i]);
        free(temp1[i]);
        free(temp2[i]);
    }
    free(ffexp);
    free(b);
    free(temp1);
    free(temp2);

    free(xatoms);
    free(yatoms);
    free(zatoms);
    free(charg2);
    free(ioffst);
    free(nbors);
    return;
}


// ******************************************************************
void addexp(double***b, double***a, int nterms) {
    int i,j,k;
    for (i=0;i<nterms+1;i++) {
        for (j=0;j<nterms+1;j++) {
            for (k=0;k<nterms+1;k++) {
                a[i][j][k] = a[i][j][k] + b[i][j][k];
            }
        }
    }
    return;
}


// ******************************************************************
void assign(int nside, double* xat, double* yat, double* zat, double*
	    str, double* xatoms, double* yatoms, double* zatoms,
	    double* charg2, int* ioffst, double* xtarg, double*
	    ytarg, double* ztarg, int ntarg, int* ibox, int natoms,
	    double* center, int* itradr, int* itofst) {
    // just to be clear, to avoid wierd 3d array effects, am leaving it
    // as a 3*nboxes length array.
    int /*i,*/ j, nboxes, ixh, iyh, izh, iadr, indx;
    double h, hh;
    int* icnt; //allocate dynamically here
    int* icnt2; // same
  
    nboxes = nside*nside*nside;
    icnt = (int*) malloc(nboxes * sizeof(int));
    icnt2 = (int*) malloc(nboxes * sizeof(int));

    for (j=0;j<nboxes;j++) {
        icnt2[j] = 0;
        icnt[j] = 0;
    }
    h = 1.0/nside;

    for (j=0;j<natoms;j++) {
        ixh = (int) floor(xat[j]/h);
        if (ixh > nside - 1) {
            ixh = nside - 1;
        }
        if (ixh < 0) {
            ixh = 0;
        }
        iyh = (int) floor(yat[j]/h);
        if (iyh > nside -1) {
            iyh = nside - 1;
        }
        if (iyh < 0) {
            iyh = 0;
        }
        izh = (int) floor(zat[j]/h);
        if (izh > nside - 1) {
            izh = nside - 1;
        }
        if (izh < 0) {
            izh = 0;
        }
        iadr = izh*nside*nside + iyh*nside + ixh;
        icnt[iadr]++;
        ibox[j] = iadr;
    }

    ioffst[0] = 0;
    for (j=1;j<nboxes+1;j++) {
        ioffst[j] = ioffst[j-1] + icnt[j-1];
    }
  
    for (j=0;j<natoms;j++) {
        iadr = ibox[j];
        indx = ioffst[iadr] + icnt2[iadr];
        xatoms[indx] = xat[j];
        yatoms[indx] = yat[j];
        zatoms[indx] = zat[j];
        charg2[indx] = str[j];
        icnt2[iadr]++;
    }

    for (j=0;j<nboxes;j++) {
        icnt2[j] = 0;
        icnt[j] = 0;
    }
    for (j=0;j<ntarg;j++) {
        ixh = (int) floor(xtarg[j]/h);
        if (ixh > nside - 1) {
            ixh = nside - 1;
        }
        if (ixh < 0) {
            ixh = 0;
        }
        iyh = (int) floor(ytarg[j]/h);
        if (iyh > nside - 1) {
            iyh = nside - 1;
        }
        if (iyh < 0) {
            iyh = 0;
        }
        izh = (int) floor(ztarg[j]/h);
        if (izh > nside - 1) {
            izh = nside - 1;
        }
        if (izh < 0) {
            izh = 0;
        }
        iadr = izh*nside*nside + iyh*nside + ixh;
        icnt[iadr]++;
        ibox[j] = iadr;
    }

    itofst[0] = 0;
    for (j = 1;j<nboxes + 1;j++) {
        itofst[j] = itofst[j-1] + icnt[j-1];
    }

    for (j=0;j<ntarg;j++) {
        iadr = ibox[j];
        indx = itofst[iadr] + icnt2[iadr];
        itradr[indx] = j;
        icnt2[iadr]++;
    }

    hh = 0.5*h;
    for (j=0;j<3*nboxes;j=j+3) {
        izh = (j/3)/(nside*nside);
        iyh = ((j/3)%(nside*nside)) / nside;
        ixh = ((j/3)%(nside*nside)) % nside;
        center[j] = ixh*h + hh;
        center[j+1] = iyh*h + hh;
        center[j+2] = izh*h + hh;
    }
    free(icnt);
    free(icnt2);
    return;
}


// ******************************************************************
void freestuff(int* iloc, int* ibox, int* itofst, int* itradr,
	       double**** locexp, double* center, int nterms, int
	       ntbox) {

    // frees all the stuff allocated dynamically in lenchk
    int i,j,k;


    free(iloc);
    free(ibox);
    free(itofst);
    free(itradr);
  
    // free locexp
    for (i=0;i<ntbox;i++) {
        for (j=0;j<(nterms+1);j++) {
            for (k=0;k<(nterms+1);k++) {
                free(locexp[i][j][k]);
            }
            free(locexp[i][j]);
        }
        free(locexp[i]);
    }
    free(locexp);

    // center
    free(center);
    return;
}


// ******************************************************************
void gshift(double*** ffexp, double* cent1, double* cent2, double***
	    local, int nterms, double*** temp1, double*** temp2, double delta) {

    int i,/*j,k,*/ len, i2, alpha1, alpha2, alpha3, beta1, beta2, beta3;
    double dsq, x, y, z, x2, y2, z2, sum, facx, facy, facz;
    double* hexpx, *hexpy, *hexpz, *fac;

    hexpx = (double*) malloc((2*nterms+1) * sizeof(double));
    hexpy = (double*) malloc((2*nterms+1) * sizeof(double));
    hexpz = (double*) malloc((2*nterms+1) * sizeof(double));
    fac = (double*) malloc((nterms+1) * sizeof(double));

    dsq = 1.0/sqrt(delta);
    x = (cent2[0]-cent1[0])*dsq;
    x2 = 2*x;
    y = (cent2[1]-cent1[1])*dsq;
    y2 = 2*y;
    z = (cent2[2]-cent1[2])*dsq;
    z2 = 2*z;

    facx = exp(-x*x);
    facy = exp(-y*y);
    facz = exp(-z*z);

    hexpx[0] = facx;
    hexpy[0] = facy;
    hexpz[0] = facz;

    hexpx[1] = x2*facx;
    hexpy[1] = y2*facy;
    hexpz[1] = z2*facz;

    for (i=1;i<(2*nterms);i++) {
        i2 = 2*i;
        hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
        hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
        hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
    }
    fac[0] = 1.0;
    for (i=1;i<nterms+1;i++) {
        fac[i] = -fac[i-1]/i;
    }

    len = nterms+1;
    for (alpha1=0;alpha1<len;alpha1++) {
        for (alpha2=0;alpha2<len;alpha2++) {
            for (beta3=0;beta3<len;beta3++) {
                sum = 0.0;
                for (alpha3=nterms;alpha3>=0;alpha3--) {
                    sum=sum+ffexp[alpha1][alpha2][alpha3]*hexpz[alpha3+beta3];
                }
                temp1[alpha1][alpha2][beta3] = sum;
            }
        }
    }

    for (alpha1=0;alpha1<len;alpha1++) {
        for (beta2=0;beta2<len;beta2++) {
            for (beta3=0;beta3<len;beta3++) {
                sum = 0.0;
                for (alpha2=nterms;alpha2>=0;alpha2--) {
                    sum=sum+temp1[alpha1][alpha2][beta3]*hexpy[alpha2+beta2];
                }
                temp2[alpha1][beta2][beta3] = sum;
            }
        }
    }

    for (beta1=0;beta1<len;beta1++) {
        for (beta2=0;beta2<len;beta2++) {
            for (beta3=0;beta3<len;beta3++) {
                sum = 0.0;
                for (alpha1=nterms;alpha1>=0;alpha1--) {
                    sum=sum+temp2[alpha1][beta2][beta3]*hexpx[alpha1+beta1];
                }
                local[beta1][beta2][beta3]=sum*fac[beta1]*fac[beta2]*fac[beta3];
            }
        }
    }
  

  /* TESTED ON August 6: the above computation is sound- it gives the same
     results as the "brute force" summation that doesn't take
     advantage of the separability of hexp */  

/*  // DEBUGGING!! */
/*    // test the stupid way */
/*    for (beta1=0;beta1<len;beta1++) */
/*      { */
/*        for (beta2=0;beta2<len;beta2++) */
/*  	{ */
/*  	  for (beta3=0;beta3<len;beta3++) */
/*  	    { */
/*  	      // begin the sum */
/*  	      sum = 0.0; */
/*  	      for (alpha1=0;alpha1<len;alpha1++) */
/*  		{ */
/*  		  for (alpha2=0;alpha2<len;alpha2++) */
/*  		    { */
/*  		      for (alpha3=0;alpha3<len;alpha3++) */
/*  			{ */
/*  			  sum = sum + */
/*  			    ffexp[alpha1][alpha2][alpha3]*hexpx[alpha1+beta1]*hexpy[alpha2+beta2]*hexpz[alpha3+beta3]; */
/*  			} */
/*  		    } */
/*  		} */
/*  	      temp1[beta1][beta2][beta3] = */
/*  		sum*fac[beta1]*fac[beta2]*fac[beta3]; */
/*  	    } */
/*  	} */
/*      } */

/*    // DEBUG!!  */
/*    // idiot checking */
/*    for (beta1=0;beta1<len;beta1++) */
/*      for (beta2=0;beta2<len;beta2++) */
/*        for (beta3=0;beta3<len;beta3++) */
/*  	printf("Local expansions: Smart way %lf Stupd way %lf\n", */
/*  	       local[beta1][beta2][beta3], */
/*  	       temp1[beta1][beta2][beta3]);			   */


    free(hexpx);
    free(hexpy);
    free(hexpz);
    free(fac);
    return;
}


// ******************************************************************
void mknbor(int ibox, int* nbors, int* nnbors, int nside, double kdis) {

    int i, j, k, ihgt, irow, icol, imin, imax, jmin, jmax, kmin, kmax;
    *nnbors = 0;
    ihgt = ibox/(nside*nside);
    irow = (ibox - ihgt*nside*nside)/nside; // same as modulo
    icol = (ibox - ihgt*nside*nside) - irow*nside; // ditto
    // threshold the x dimension
    if ((icol-kdis)>0)
        imin = (int) floor(icol-kdis);
    else
        imin = 0;
    if ((icol+kdis)<(nside-1))
        imax = (int) floor(icol + kdis);
    else
        imax = nside-1;
    // threshold y
    if ((irow-kdis)>0)
        jmin = (int) floor(irow - kdis);
    else
        jmin = 0;
    if ((irow+kdis)<(nside-1))
        jmax = (int) floor(irow+kdis);
    else 
        jmax = nside-1;
    // threshold z
    if ((ihgt-kdis)>0)
        kmin = (int) floor(ihgt-kdis);
    else
        kmin = 0;
    if ((ihgt+kdis)<(nside-1))
        kmax = (int) floor(ihgt+kdis);
    else
        kmax = nside-1;

    for (i=imin;i<imax+1;i++) {
        for (j=jmin;j<jmax+1;j++) {
            for (k=kmin;k<kmax+1;k++) {
                nbors[*nnbors] = k*nside*nside+j*nside+i;
                (*nnbors)++;
            }
        }
    }
    return;
}


// ******************************************************************
void lenchk(int nside, double* xtarg, double* ytarg, double* ztarg,
	    int ntarg, int nterms, int nmax, int** iloc, int** ibox,
	    int** itofst, int** itradr, double***** locexp, double**
	    center, int* ntbox) {

    int i, j, k, nboxes, ixh, iyh, izh, iadr;
    double h;

    nboxes = nside*nside*nside;
    *ntbox = 0;
    // allocate the arrays
    *iloc = (int*) malloc(nboxes * sizeof(int));
    *itofst = (int*) malloc((nboxes+1) * sizeof(int));
    *itradr = (int*) malloc(ntarg * sizeof(int));
    *ibox = (int*) malloc(nmax * sizeof(int));

    // allocating the locexp array AFTER calculating ntbox!

    // centers 
    *center = (double*) malloc(3*nboxes * sizeof(double));

    h = 1.0/nside;
    for (i=0;i<nboxes;i++) {
        (*iloc)[i] = 0;
    }

    // for each target point...
    for (j=0;j<ntarg;j++) {
        // determine which box this point will go into.
        ixh = (int) floor(xtarg[j]/h);
        iyh = (int) floor(ytarg[j]/h);
        izh = (int) floor(ztarg[j]/h);
        if (ixh > nside -1)
            ixh = nside - 1;
        if (ixh < 0)
            ixh = 0;
        if (iyh > nside - 1)
            iyh = nside - 1;
        if (iyh < 0)
            iyh = 0;
        if (izh > nside - 1)
            izh = nside - 1;
        if (izh < 0)
            izh = 0;
        // iadr is the index of the box to which this point is assigned.
        iadr = izh*nside*nside + iyh*nside + ixh;
        if ((*iloc)[iadr]==0) {
            (*iloc)[iadr] = *ntbox;
            // number of target boxes?
            (*ntbox) = (*ntbox) + 1;
        }
    }


    // allocating the locexp array
    *locexp = (double****) malloc(*ntbox * sizeof(double***));
    for(i = 0; i < *ntbox; i++) {
        (*locexp)[i] = (double***) malloc((nterms+1) * sizeof(double**));
        for (j=0; j<(nterms+1); j++) {
            (*locexp)[i][j] = (double**) malloc((nterms+1) * sizeof(double*));
            for (k=0;k<(nterms+1);k++) {
                (*locexp)[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
            }
        }
    }
    return;
}

double scale_to_cube(double* sx, double* sy, double* sz, int NX, double* tx, double* ty,
                   double* tz, int NY) {
    int i;
    double maxx = -INF;
    double minx = INF;
    double maxy = -INF;
    double miny = INF;
    double maxz = -INF;
    double minz = INF;
    double dx, dy, dz;
    double scale;
    for (i = 0; i < NX; i++) {
        maxx = dmax(maxx, sx[i]);
        minx = dmin(minx, sx[i]);
        maxy = dmax(maxy, sy[i]);
        miny = dmin(miny, sy[i]);
        maxz = dmax(maxz, sz[i]);
        minz = dmin(minz, sz[i]);
    }
    for (i = 0; i < NY; i++) {
        maxx = dmax(maxx, tx[i]);
        minx = dmin(minx, tx[i]);
        maxy = dmax(maxy, ty[i]);
        miny = dmin(miny, ty[i]);
        maxz = dmax(maxz, tz[i]);
        minz = dmin(minz, tz[i]);
    }
    dx = maxx - minx;
    dy = maxy - miny;
    dz = maxz - minz;
    scale = dmax(dx, dy);
    scale = dmax(scale, dz);
    for (i = 0; i < NX; i++) {
        sx[i] = (sx[i] - minx) / scale;
        sy[i] = (sy[i] - miny) / scale;
        sz[i] = (sz[i] - minz) / scale;
    }
    for (i = 0; i < NY; i++) {
        tx[i] = (tx[i] - minx) / scale;
        ty[i] = (ty[i] - miny) / scale;
        tz[i] = (tz[i] - minz) / scale;
    }
    return scale;
}

void gaussth(double *X, double *Y, double *w, double* kparams, int NP, char* ker_type,
             int NX, int NY, int D, double* bests, double eps_abs) {
    double* sx;
    double* sy;
    double* sz;
    double* tx;
    double* ty;
    double* tz;
    double h;
    double scale;
    int i;
    if (strcmp(ker_type, "gaussian") != 0) {
        printf("fgt only works with gaussian kernels\n");
        return;
    }
    if (D > 3) {
        printf("fgt doesn't work for D > 3\n");
        return;
    }
    h = kparams[0];
    sx = (double*) malloc(NX * sizeof(double));
    sy = (double*) malloc(NX * sizeof(double));
    sz = (double*) malloc(NX * sizeof(double));
    tx = (double*) malloc(NY * sizeof(double));
    ty = (double*) malloc(NY * sizeof(double));
    tz = (double*) malloc(NY * sizeof(double));
    for (i = 0; i < NX; i++) {
        sx[i] = X[i*D];
        if (D >= 2) {
            sy[i] = X[i*D + 1];
        } else {
            sy[i] = 0;
        }
        if (D >= 3) {
            sz[i] = X[i*D + 2];
        } else {
            sz[i] = 0;
        }
    }
    for (i = 0; i < NY; i++) {
        tx[i] = Y[i*D];
        if (D >= 2) {
            ty[i] = Y[i*D + 1];
        } else {
            ty[i] = 0;
        }
        if (D >= 3) {
            tz[i] = Y[i*D + 2];
        } else {
            tz[i] = 0;
        }
    }
    scale = scale_to_cube(sx, sy, sz, NX, tx, ty, tz, NY);
    gausst(sx, sy, sz, w, NX, tx, ty, tz, bests, NY, h*h/(scale*scale), eps_abs, 0);
    free(sx);
    free(sy);
    free(sz);
    free(tx);
    free(ty);
    free(tz);
}
