/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef GENERIC_H_
#define GENERIC_H_

#include "kernel.h"

struct tree_globals {
  int D; // data is in R^D
  double* X; // source particles
  double* Y; // target particles
  double* w; // weight for each source particle
  int* ibest; // for each target particle, index of the most influential source particle
  double* best; // for each target particle, estimate of the sum of weighted influences
  int influence_mode; // POSITIVE_INFLUENCE, NEGATIVE_INFLUENCE or MIXED_INFLUENCE
  int weights_mode; // POSITIVE_WEIGHTS, NEGATIVE_WEIGHTS or MIXED_WEIGHTS
};
typedef struct tree_globals tree_globals;

struct Node {
    /** The number of points contained by this node. */
    int N;
    /** The indices of points contained by this node. */
    int* indices;
    /** The number of children this node has. */
    int Nchildren;
    /** The children of this node. */
    struct Node** children;
    // only ever need one of these!
//    union {
        double sumposweight;
        double sumnegweight;
        double maxweight;
        double minweight;  // in case influence is negative!
//    };
    void* extra;
};
typedef struct Node Node;

/************** Node-specific *************/
// These functions must be implemented by the structure used
/* Returns a lower bound on the distance between two nodes. */
double (*node_node_mindist)(Node* xnode, Node* ynode, tree_globals* globals);
/* Returns an upper bound on the distance between two nodes. */
double (*node_node_maxdist)(Node* xnode, Node* ynode, tree_globals* globals);
// returns the minimum distance between a (source) node and a (target) particle
double (*node_particle_mindist)(Node* xnode, double* yparticle, tree_globals* globals);
// returns the maximum distance between a (source) node and a (target) particle
double (*node_particle_maxdist)(Node* xnode, double* yparticle, tree_globals* globals);
/******************************************/

/* Given a node in your data structure, is it a leaf node? */
int is_leaf(Node* node);
/* Given a node, how many particles does its contain? */
//int get_N(Node* node);

/* Get the global ids for the particles at this node in the X or Y tree. */
int* get_indices(Node* node);
//int* get_x_indices(Node* node, tree_globals* globals);

double get_weight(Node* node, int indx, tree_globals* globals);

/* Return the maximum weighted particle in this subtree. */
double get_sumposweight(Node* node);
double get_sumnegweight(Node* node);
/* Return the number of children of this node. */
//int  N_children(Node* node);
/* Get the "c"th child of this node. */
Node* get_child(Node* vnode, int c);

/* This function is called when the answer has been computed.
 * For the "index"th particle at the given Y node, the argmax
 * ("bestindex") and max ("bestinfluence") are given.  Chances are
 * that you just want to store this value. */
// update the global 'bests':
// for max these are the indices of the most influential source particles, one for each target particle
// for sum these are the estimates of the sum for each target particle
void set_best_influence(Node* ynode, int indx, int bestindex, double bestinfluence,
                         tree_globals* globals);

/* The following are used only by "generic_decorate_x_tree",
   which does not really have to be used - it's just a convenience
   for computing the "maxweights" and the weight-sorted indices. */

/* Get the identifier of the "index"th particle at this node. */
int get_unsorted_index(Node* node, int indx);
///* Set the maximum weight in this subtree. */
void set_sumposweight(Node* node, double sumposweight);
void set_sumnegweight(Node* node, double sumnegweight);

/************ The following are used by "generic_max" ***********/

double* get_y_particle(Node* node, int i, tree_globals* globals);
double* get_x_particle(Node* node, int i, tree_globals* globals);

double get_maxweight(Node* node);
double get_minweight(Node* node);

void set_sorted_indices(Node* xnode, int* sorted);
void set_maxweight(Node* xnode, double maxweight);
void set_minweight(Node* node, double minweight);

/*************** General ***************/

void free_tree(Node* tree, void (*free_extra)(Node* node));

#endif /*GENERIC_H_*/
