/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "generic_max.h"
#include "nodelist.h"

/**
 *
 * (Somewhat) Data-structure generic tree-based max-weighted-distance-function code.
 *
 * The assumptions:
 *
 *   -we wish to find max and argmax of a weighted function of distance
 *    between points in space.  (And the space has some notion of distance
 *    - and the triangle inequality?)
 *
 *   -there are many "sources" (kernels, particles, masses, whatever) and many
 *    point at which we want to find the maximum influence.
 *
 *   -the function is monotonic decreasing (ideally, rapidly so)
 *
 *   -there is a tree-like data structure that splits the points into a
 *    hierarchy of clusters.
 *
 *   -the data structure allows the distance between two clusters and
 *    between a point and all the points in a cluster to be bounded
 *    (ideally, cheaply).
 *
 *   -the data structure stores, at each node, indices into an externally-stored
 *    array of points. [actually, no longer true, but there is still some bias
 *    toward this type of storage.  it's probably a good idea, anyway.]
 *
 *
 * The data structures:
 *
 *   -almost everything is treated as an opaque void* structure, and accessor
 *    functions are used to get at things.
 *
 *   -a void* "parameter" is used to pass around global constants and such.
 *
 *
 * The inputs:
 *
 *   -a "distance" function that takes two void* arguments and returns the
 *    distance (or some monotonic increasing function of distance like the
 *    distance squared)
 *
 *   -functions that allow the maximum weight in a subtree to be stored and
 *    retrieved.
 */

/**
   As we recurse down the trees, we add potential X nodes into the
   "leaflist" list.

   Meanwhile, "Xnodes" contains non-leaf X nodes.

   Once the parent list is empty, we can filter the "leaflist" list
   by the final best box value, rejecting any that no longer qualify.
**/

INLINE double miub_mwmi(Node* xnode, Node* ynode, tree_globals* globals) {
    double d1 = node_node_maxdist(xnode, ynode, globals);
    double d2 = node_node_mindist(xnode, ynode, globals);
    double i1 = dmax(kernel_influence(get_maxweight(xnode), d1), kernel_influence(get_minweight(xnode), d1));
    double i2 = dmax(kernel_influence(get_maxweight(xnode), d2), kernel_influence(get_minweight(xnode), d2));
    return dmax(i1,i2);
}

INLINE double milb_mwmi(Node* xnode, Node* ynode, tree_globals* globals) {
    double d1 = node_node_maxdist(xnode, ynode, globals);
    double d2 = node_node_mindist(xnode, ynode, globals);
    double i1 = dmin(kernel_influence(get_maxweight(xnode), d1), kernel_influence(get_minweight(xnode), d1));
    double i2 = dmin(kernel_influence(get_maxweight(xnode), d2), kernel_influence(get_minweight(xnode), d2));
    return dmin(i1,i2);
}

INLINE double miubp_mwmi(Node* xnode, double* yp, tree_globals* globals) {
    double d1 = node_particle_maxdist(xnode, yp, globals);
    double d2 = node_particle_mindist(xnode, yp, globals);
    double i1 = dmax(kernel_influence(get_maxweight(xnode), d1), kernel_influence(get_minweight(xnode), d1));
    double i2 = dmax(kernel_influence(get_maxweight(xnode), d2), kernel_influence(get_minweight(xnode), d2));
    return dmax(i1,i2);
}

INLINE double max_influence_upper_bounds(Node* xnode, Node* ynode, tree_globals* globals) {
    double dist = 0;
    if (globals->weights_mode == POSITIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == POSITIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_node_maxdist(xnode, ynode, globals);
        } else {
            dist = node_node_mindist(xnode, ynode, globals);
        }
    } else if (globals->weights_mode == NEGATIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == NEGATIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_node_mindist(xnode, ynode, globals);
        } else {
            dist = node_node_maxdist(xnode, ynode, globals);
        }
    }
    if (globals->influence_mode == POSITIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == POSITIVE_WEIGHTS)) {
        return kernel_influence(get_maxweight(xnode), dist);
    } else if (globals->influence_mode == NEGATIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == NEGATIVE_WEIGHTS)) {
        return kernel_influence(get_minweight(xnode), dist);
    }
    if (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == MIXED_INFLUENCE) {
        return miub_mwmi(xnode, ynode, globals);
    }
    return -INF;
}

INLINE double max_influence_lower_bounds(Node* xnode, Node* ynode, tree_globals* globals) {
    double dist = 0;
    if (globals->weights_mode == POSITIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == POSITIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_node_mindist(xnode, ynode, globals);
        } else {
            dist = node_node_maxdist(xnode, ynode, globals);
        }
    } else if (globals->weights_mode == NEGATIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == NEGATIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_node_maxdist(xnode, ynode, globals);
        } else {
            dist = node_node_mindist(xnode, ynode, globals);
        }
    }
    if (globals->influence_mode == POSITIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == POSITIVE_WEIGHTS)) {
        return kernel_influence(get_maxweight(xnode), dist);
    } else if (globals->influence_mode == NEGATIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == NEGATIVE_WEIGHTS)) {
        return kernel_influence(get_minweight(xnode), dist);
    }
    if (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == MIXED_INFLUENCE) {
        return milb_mwmi(xnode, ynode, globals);
    }
    return INF;
}

INLINE double max_influence_upper_bounds_p(Node* xnode, double* yp, tree_globals* globals) {
    double dist = 0;
    if (globals->weights_mode == POSITIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == POSITIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_particle_maxdist(xnode, yp, globals);
        } else {
            dist = node_particle_mindist(xnode, yp, globals);
        }
    } else if (globals->weights_mode == NEGATIVE_WEIGHTS ||
        (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == NEGATIVE_INFLUENCE)) {
        if (kernel_increasing()) {
            dist = node_particle_mindist(xnode, yp, globals);
        } else {
            dist = node_particle_maxdist(xnode, yp, globals);
        }
    }
    if (globals->influence_mode == POSITIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == POSITIVE_WEIGHTS)) {
        return kernel_influence(get_maxweight(xnode), dist);
    } else if (globals->influence_mode == NEGATIVE_INFLUENCE ||
        (globals->influence_mode == MIXED_INFLUENCE && globals->weights_mode == NEGATIVE_WEIGHTS)) {
        return kernel_influence(get_minweight(xnode), dist);
    }
    if (globals->weights_mode == MIXED_WEIGHTS && globals->influence_mode == MIXED_INFLUENCE) {
        return miubp_mwmi(xnode, yp, globals);
    }
    return -INF;
}

INLINE double translate_influence(double infl, double maxweight, double weight) {
    return infl / maxweight * weight;
}

void generic_max_recurse(Node* Ynode, Node** Xnodes, int Nxnodes, double parentbest,
                         nodelist* leaflist, tree_globals* globals) {
    if (is_leaf(Ynode) && Nxnodes == 0) {
        int xbox, x, y;
        int Nleaves;
        double* maxinfls;
        double* mininfls;
        int* order;
        int Nkeep, n;
        double best;
    
        // note that we cannot mutate the leaflist in this function, since it will be
        // used by our siblings in the 'Y' tree.
    
        // to work around this, we cache the maxinfls for each x node to the y node.
        // we also compute the mininfls and sort the leaves by this order, so that
        // we (hopefully) find the best "best" as soon as possible.
        Nleaves = nodelist_size(leaflist);
        maxinfls = (double*) malloc(Nleaves * sizeof(double));
        mininfls = (double*) malloc(Nleaves * sizeof(double));
        order = (int*) malloc(Nleaves * sizeof(int));
        Nkeep = 0;
        best = parentbest;
        for (x = 0; x < Nleaves; x++) {
            Node* xnode = nodelist_get(leaflist, x);
            maxinfls[x] = max_influence_upper_bounds(xnode, Ynode, globals);
            if (maxinfls[x] >= best) {
                // it is possible that the source leaf contains the particle we want
                mininfls[x] = max_influence_lower_bounds(xnode, Ynode, globals);
            	if (mininfls[x] > best) {
                    best = mininfls[x];
                }
                order[Nkeep++] = x;
            }
        }

        // another round of pruning, with the new 'best'.
        n = 0;
        for (x = 0; x < Nkeep; x++) {
            if (maxinfls[order[x]] >= best) {
                order[n++] = order[x];
            }
        }
        Nkeep = n;
        parentbest = best;
    
        // sort the order.
        compare_weights_vector = mininfls;
        qsort(order, Nkeep, sizeof(int), compare_weights);
        compare_weights_vector = NULL;
    
        // look at individual Y particles...
        for (y = 0; y < Ynode->N; y++) {
            double best;
            int ibest;
            int o;
            double* yparticle;
    
            ibest = -2;
            best = parentbest;
            yparticle = get_y_particle(Ynode, y, globals);
    
            for (o = 0; o < Nkeep; o++) {
                Node* Xnode;
    	        double maxinfl, maxweight, minweight;
    	        int NX;
    
    	        xbox = order[o];
    	        Xnode = nodelist_get(leaflist, xbox);
    
                // see if we can skip this node based on its cached maxinfl val.
                if (maxinfls[xbox] < best) {
                    // we've already found a better particle than any particle in this source leaf
                    continue;
            	}
    
                // see if we can skip this box entirely...
                maxinfl = max_influence_upper_bounds_p(Xnode, yparticle, globals);
    	        maxweight = get_maxweight(Xnode);
                minweight = get_minweight(Xnode);
    
            	if (maxinfl < best) {
                    // we've already found a better particle than any particle in this source leaf
                    continue;
            	}
    	
                NX = Xnode->N;

                for (x = 0; x < NX; x++) {
    	            double dist;
                    double infl;
                    double weight;
                    double* xparticle;
                    int tx = x;
                    if (globals->influence_mode == NEGATIVE_INFLUENCE) {
                        tx = NX - x - 1;
                    }
            	    // particles are sorted by weight - with the mindist, can we ignore the
            	    // remainder of the particles?
                    // weight = get_weight(Xnode, x, globals);
                    weight = get_weight(Xnode, tx, globals);
                    if (translate_influence(maxinfl, maxweight, weight) < best &&
                        translate_influence(maxinfl, minweight, weight) < best) {
    	                // skip the rest
                        if (globals->influence_mode == MIXED_INFLUENCE) {
                            continue;
                        } else {
                            break;
                        }
                    }
                    xparticle = get_x_particle(Xnode, tx, globals);
                    dist = kernel_distance(xparticle, yparticle, globals->D);
                    infl = kernel_influence(weight, dist);
                    if (ibest < 0) {
                        if (infl - best > -1e-15) {
                    	    best = infl;
                    	    ibest = get_unsorted_index(Xnode, tx);
                        }
                    } else {
                        if (infl >= best) {
                            best = infl;
                            ibest = get_unsorted_index(Xnode, tx);
                        }
                    }
                }
            }
            set_best_influence(Ynode, y, ibest, best, globals);
        }
    
        free(maxinfls);
        free(mininfls);
        free(order);
    } else {
        int c;
        int x;
        int NXC;
        Node** Xkeepers;
        double *maxinfls, *mininfls;
        int Nyrecurse;
        int nleaves_orig;

        // count number of children of non-leaf X nodes.
        NXC = 0;
        for (x = 0; x < Nxnodes; x++) {
            NXC += Xnodes[x]->Nchildren;
        }
    
        Xkeepers = (Node**) malloc(NXC * sizeof(Node*));
        maxinfls = (double*) malloc(NXC * sizeof(double));
        mininfls = (double*) malloc(NXC * sizeof(double));
        
        if (is_leaf(Ynode)) {
            Nyrecurse = 1;
        } else {
            Nyrecurse = Ynode->Nchildren;
        }

        nleaves_orig = nodelist_size(leaflist);

        // recurse on Y nodes...
        for (c = 0; c < Nyrecurse; c++) {
            Node* Ychild;
            double best;
            int Nkeepers = 0;
            int nkeep;
            int x;
			int* order;
			int i;

            if (is_leaf(Ynode)) {
                Ychild = Ynode;
            } else {
	            Ychild = get_child(Ynode, c);
            }
            best = parentbest;
            
            for (x = 0; x < Nxnodes; x++) {
                int Nx, xc;
                double mininfl, maxinfl;
                Node* xnode = Xnodes[x];

	            // check if we can reject the parent box...
                maxinfl = max_influence_upper_bounds(xnode, Ychild, globals);
                if (maxinfl < best) {
                    // this node is not promising
                    continue;
                }

                Nx = xnode->Nchildren;

                for (xc = 0; xc < Nx; xc++) {
                    Node* Xchild = get_child(xnode, xc);
                    // check to see if we can ignore this X child...
                    maxinfl = max_influence_upper_bounds(Xchild, Ychild, globals);
                    if (maxinfl < best) {
                        // not a promising child
                        continue;
                    }
                    // check to see if this X child is the new best...
                    mininfl = max_influence_lower_bounds(Xchild, Ychild, globals);
                    if (mininfl > best) {
                        // very promising child
                        best = mininfl;
                    }
                    // mark this child as a potential keeper...
                    maxinfls[Nkeepers] = maxinfl;
                    mininfls[Nkeepers] = mininfl;
                    Xkeepers[Nkeepers] = Xchild;
                    Nkeepers++;
                }
            }

            // recurse on this Y child:

            // -reject any potential keepers that fall below the final "best" val.
            nkeep = 0;
            for (x = 0; x < Nkeepers; x++) {
                if (maxinfls[x] < best) {
                    // not that great after all
                    continue;
                } else {
                    if (is_leaf(Xkeepers[x])) {
                        // keep this leaf
                        nodelist_append(leaflist, Xkeepers[x]);
                    } else {
                        Xkeepers[nkeep] = Xkeepers[x];
                        mininfls[nkeep] = mininfls[x];
                        nkeep++;
                    }
                }
            }
            Nkeepers = nkeep;

            // -sort the keepers by their mininfls.
            order = (int*) malloc(Nkeepers * sizeof(int));
            for (i = 0; i < Nkeepers; i++) {
                order[i] = i;
            }
            compare_weights_vector = mininfls;
            qsort(order, Nkeepers, sizeof(int), compare_weights);
            compare_weights_vector = NULL;

            permute_nodes(Xkeepers, order, Nkeepers);
            free(order);
      
            generic_max_recurse(Ychild, Xkeepers, Nkeepers, best, leaflist, globals);

            // remove potential leaves that we added for this child.
            nodelist_remove_after_index(leaflist, nleaves_orig);
        }
        free(Xkeepers);
        free(maxinfls);
        free(mininfls);
    }
}

void generic_max(tree_globals* globals, Node* Xroot, Node* Yroot) {
    nodelist* xleaves = nodelist_new();
    // if the root of the X tree is also a leaf, then add it to the xleaves list.
    if (is_leaf(Xroot)) {
        nodelist_append(xleaves, Xroot);
        generic_max_recurse(Yroot, NULL, 0, -INF, xleaves, globals);
    } else {
        generic_max_recurse(Yroot, &Xroot, 1, -INF, xleaves, globals);
    }
    nodelist_free(xleaves);
}

INLINE void generic_max_decorate_x_tree(Node* node, double* w, tree_globals* globals) {
    double maxw = -INF;
    double minw = INF;
    if (is_leaf(node)) {
        int* sorted;
        int i, N;        
        N = node->N;
        
        sorted = (int*) malloc(N * sizeof(int));
        for (i = 0; i < N; i++) {
            sorted[i] = get_unsorted_index(node, i);
        }
                
        compare_weights_vector = w;
        qsort(sorted, N, sizeof(int), compare_weights);
        set_sorted_indices(node, sorted);
        maxw = w[sorted[0]];
        minw = w[sorted[N-1]];
    } else {
        int NC;
        int c;
        
        NC = node->Nchildren;
        
        for (c = 0; c < NC; c++) {
            Node* child = get_child(node, c);
            generic_max_decorate_x_tree(child, w, globals);
            maxw = dmax(maxw, get_maxweight(child));
            minw = dmin(minw, get_minweight(child));
        }
    }
    set_maxweight(node, maxw);
    set_minweight(node, minw);
}
