/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "generic_sum.h"
#include "pqueue.h"

pqueue* pq;

int hopeless;

int num_satisfied;

INLINE void sib_pw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower) {
    double sumposweight = get_sumposweight(xnode);
    double mindist = node_node_mindist(xnode, ynode, globals);
    double maxdist = node_node_maxdist(xnode, ynode, globals);
//    if (kernel_increasing() == -1) {
//        double d1 = kernel_influence(sumposweight, maxdist);
//        double d2 = kernel_influence(sumposweight, mindist);
//        *upper = dmax(d1,d2);
//        *lower = dmin(d1,d2);
//    } else
    if (kernel_increasing()) {
        *upper = kernel_influence(sumposweight, maxdist);
        *lower = kernel_influence(sumposweight, mindist);
    } else {
        *upper = kernel_influence(sumposweight, mindist);
        *lower = kernel_influence(sumposweight, maxdist);    
    }
}

INLINE void sib_nw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower) {
    double sumnegweight = get_sumnegweight(xnode);
    double mindist = node_node_mindist(xnode, ynode, globals);
    double maxdist = node_node_maxdist(xnode, ynode, globals);
//    if (kernel_increasing() == -1) {
//        double d1 = kernel_influence(sumnegweight, maxdist);
//        double d2 = kernel_influence(sumnegweight, mindist);
//        *upper = dmax(d1,d2);
//        *lower = dmin(d1,d2);
//    } else 
    if (kernel_increasing()) {
        *upper = kernel_influence(sumnegweight, mindist);
        *lower = kernel_influence(sumnegweight, maxdist);
    } else {
        *upper = kernel_influence(sumnegweight, maxdist);
        *lower = kernel_influence(sumnegweight, mindist);    
    }
}

INLINE void sib_mw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower) {
    double sumposweight = get_sumposweight(xnode);
    double sumnegweight = get_sumnegweight(xnode);
    double mindist = node_node_mindist(xnode, ynode, globals);
    double maxdist = node_node_maxdist(xnode, ynode, globals);
//    if (kernel_increasing() == -1) {
//        double d1 = kernel_influence(sumposweight, maxdist) + kernel_influence(sumnegweight, mindist);
//        double d2 = kernel_influence(sumposweight, mindist) + kernel_influence(sumnegweight, maxdist);
//        *upper = dmax(d1,d2);
//        *lower = dmin(d1,d2);
//    } else 
    if (kernel_increasing()) {
        *upper = kernel_influence(sumposweight, maxdist) + kernel_influence(sumnegweight, mindist);
        *lower = kernel_influence(sumposweight, mindist) + kernel_influence(sumnegweight, maxdist);
    } else {
        *upper = kernel_influence(sumposweight, mindist) + kernel_influence(sumnegweight, maxdist);
        *lower = kernel_influence(sumposweight, maxdist) + kernel_influence(sumnegweight, mindist);    
    }
}

INLINE void sum_influence_bounds(Node* xnode, Node* ynode, tree_globals* globals, double* upper,
                              double* lower) {
    if (xnode->N == 1 && ynode->N == 1) {
        double dist = kernel_distance(get_x_particle(xnode, 0, globals),
                                       get_y_particle(ynode, 0, globals), globals->D);
        double infl = kernel_influence(get_weight(xnode,0,globals), dist);
        *upper = infl;
        *lower = infl;
    } else {
        if (globals->weights_mode == POSITIVE_WEIGHTS) {
            sib_pw(xnode, ynode, globals, upper, lower);
        } else if (globals->weights_mode == NEGATIVE_WEIGHTS) {
            sib_nw(xnode, ynode, globals, upper, lower);
        } else if (globals->weights_mode == MIXED_WEIGHTS) {
            sib_mw(xnode, ynode, globals, upper, lower);
        }        
    }
}

INLINE int get_order(double value) {
    return (int) floor(log(value));
}

INLINE int check_error(double error, double estimate, double eps_abs, double eps_rel) {
    error = dabs(error);
    if (estimate > 0) {
        estimate = dmax(estimate - error, 0);
    } else {
        estimate = dabs(dmin(estimate + error, 0));
    }
    if (error <= eps_abs || error <= estimate * eps_rel) {
        return 1;
    } else {
        return 0;
    }
}

INLINE
int propagate_bounds(Node* Ychild, double* estimates, double* bounds,
                      int* satisfied, double dest, double derr,
                      double epsilon_absolute, double epsilon_relative) {
    // propagate the new error bounds to all the particles of Ychild.
    int i, N;
    int* inds;
	int nsat = 0;

    N = Ychild->N;
    inds = get_indices(Ychild);
    for (i=0; i<N; i++) {
        int ind;
        ind = inds[i];
        estimates[ind] += dest;
        bounds[ind] += derr;
        if (check_error(bounds[ind], estimates[ind], epsilon_absolute,
                                     epsilon_relative)) {
            if (!satisfied[ind]) {
                satisfied[ind] = 1;
                num_satisfied++;
            }
            nsat++;
        }
    }
    if (nsat == N) {
        // all done here!
        return 0;
    }
    return 1;
}

/**
 *  I call the "data" tree X and the "query" tree Y.
 *
 * The idea:
 *   -for each particle in Y, keep track of the upper and lower bounds.
 *   -keep a list of node-node pairs (frontiers) in X and Y trees.
 *   -while the frontier list is non-empty,
 *      -remove the frontier with the largest error (upper - lower bound)
 *      -add the (children of x) * (children of y) to the frontier list.
 *      -calculate the new bounds
 *      -propagate the adjusted bounds to all of the children of y.
 *      -check the children of y to see if their new errors satisfy the error target.
 *         -if so, remove all the frontier nodes involving that y node.
 *         -propagate upward: if any of the parents of the empty node have all their
 *            children satisfied, then remove all frontier nodes involving them.
 *
 */
INLINE 
void generic_sum_helper(tree_globals* globals, double epsilon_relative, double epsilon_absolute,
                         double* estimates, double* bounds, int* satisfied, int NY) {

    Node* Ynode; // target particle node
    Node* Xnode; // source particle node
    bound_node* bnode;

    while (1) {
		int NYC, NXC;
        int xc, yc;
		int is_xleaf, is_yleaf;

        bnode = pq_pop(pq);
        
        if (bnode == NULL) {
            /* printf("no more bnodes\n"); */
            break;
        }

        // process this bound_node

        Ynode = bnode->Ynode;
        Xnode = bnode->Xnode;

        is_yleaf = is_leaf(Ynode);
        is_xleaf = is_leaf(Xnode);           
        NYC = (is_yleaf ? 1 : Ynode->Nchildren);
        NXC = (is_xleaf ? 1 : Xnode->Nchildren);

        // for Y children...
        for (yc = 0; yc < NYC; yc++) {
            Node* Ychild;
            double dest, derr;
			double NN;
			bound_node** newnodes;
			int numnew;
            
            Ychild = (is_yleaf ? Ynode : get_child(Ynode, yc));
            NN = Ychild->N;
        
            // for this Y child, calculate the change in upper and lower bounds
            // (= -old bound + sum(new bounds) )
            dest = -(bnode->upper + bnode->lower) / 2.0;
            derr = -(bnode->upper - bnode->lower) / 2.0;

            newnodes = (bound_node**) malloc(NXC * sizeof(bound_node*));
            numnew = 0;
            for (xc = 0; xc < NXC; xc++) {
                Node* Xchild;
                bound_node* newnode = (bound_node*) malloc(sizeof(bound_node));
                Xchild = (is_xleaf ? Xnode : get_child(Xnode, xc));
                newnode->Xnode = Xchild;
                newnode->Ynode = Ychild;

                // calculate new bounds.
                sum_influence_bounds(Xchild, Ychild, globals, &(newnode->upper), &(newnode->lower));
                
                // for each X child, record the change in upper and lower bounds.
                dest += (newnode->upper + newnode->lower) / 2.0;
                derr += (newnode->upper - newnode->lower) / 2.0;

                if (is_leaf(Xchild) && is_leaf(Ychild)) {
                    free(newnode);
                } else {
                    newnode->value = NN * (newnode->upper - newnode->lower);
                    newnode->order = get_order(newnode->value);
                    if (newnode->order > hopeless) {
                        newnodes[numnew++] = newnode;
                    } else {
                        free(newnode);
                    }
                }
            }
            if (propagate_bounds(Ychild, estimates, bounds, satisfied, dest, derr,
                             epsilon_absolute, epsilon_relative)) {
                // add newnodes!
                int i;
                for (i = 0; i < numnew; i++) {
                    pq_enqueue(pq, newnodes[i]);
                }
            } else {
                int i;
                for (i = 0; i < numnew; i++) {
                    free(newnodes[i]);
                }
            }
			free(newnodes);
        }

        // finished the node, free it
        free(bnode);
        if (num_satisfied == NY) {
            break;
        }
        
    }
}

// estimate the sum of source particle influences on each target particle to the given tolerances
INLINE void generic_sum(tree_globals* globals, Node* Xroot, Node* Yroot, double epsilon_relative,
         double epsilon_absolute) {
    int i, N;
    bound_node* bnode = (bound_node*) malloc(sizeof(bound_node));
    double *estimates, *bounds;
    int *satisfied;
	double est, err;
	int* inds;

    hopeless = get_order(epsilon_absolute / (double) Yroot->N);
    num_satisfied = 0;

    N = Yroot->N;
    estimates = (double*) malloc(N * sizeof(double));
    bounds = (double*) malloc(N * sizeof(double));
    satisfied = (int*) malloc(N * sizeof(int));

    // set up the bound_node for the root
    bnode->Xnode = Xroot;
    bnode->Ynode = Yroot;
    sum_influence_bounds(Xroot, Yroot, globals, &(bnode->upper), &(bnode->lower));
    bnode->value = (bnode->upper - bnode->lower) * N;
    bnode->order = get_order(bnode->value);
    
    pq = pq_new(bnode->order, hopeless);

    if (bnode->order > hopeless) {    
        pq_enqueue(pq, bnode);
    }
    est = (bnode->upper + bnode->lower) / 2.0;
    err = (bnode->upper - bnode->lower) / 2.0;
    for (i = 0; i < N; i++) {
        satisfied[i] = 0;
        estimates[i] = est;
        bounds[i] = err;
    }

    // do all the work!
    generic_sum_helper(globals, epsilon_relative, epsilon_absolute, estimates, bounds, satisfied, N);
    // give all the results
    inds = get_indices(Yroot);
    for (i = 0; i < N; i++) {
        set_best_influence(Yroot, i, -1, estimates[inds[i]], globals);
    }
    
    pq_free(pq);

    free(estimates);
    free(bounds);
    free(satisfied);
}

// decorate the tree by filling in the sumweights
INLINE void generic_sum_decorate_x_tree(Node* node, double* w, tree_globals* globals) {
    double sumpw = 0;
    double sumnw = 0;
    if (is_leaf(node)) {
        int N = node->N;
        int i;
        for (i = 0; i < N; i++) {
            double wt = w[get_unsorted_index(node, i)];
            if (wt > 0) {
                sumpw += wt;
            } else {
                sumnw += wt;
            }
        }
    } else {
        int NC = node->Nchildren;
        int c;        
        for (c = 0; c < NC; c++) {
            Node* child = get_child(node, c);
            generic_sum_decorate_x_tree(child, w, globals);
            sumpw += get_sumposweight(child);
            sumnw += get_sumnegweight(child);
        }
    }
    set_sumposweight(node, sumpw);
    set_sumnegweight(node, sumnw);
}
