/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "generic.h"
#include "kernel_gaussian.h"
#include "kernel_laplacian.h"
#include "kernel_lpq.h"
#include "kernel_sigmoid.h"
#include "kernel_polynomial.h"
#include "kernel_dog.h"
#include "kernel_gaussian_neg.h"

INLINE double dmax(double a, double b) {
  return (a > b) ? a : b;
}

INLINE double dmin(double a, double b) {
  return (a < b) ? a : b;
}

INLINE double dabs(double d) {
  return (d > 0) ? d : -d;
}

INLINE double square(double d) {
    return d * d;
}

#ifdef WINDOWS
INLINE double drand48() {
    return rand() / (double) RAND_MAX;
}

INLINE void srand48(unsigned int l) {
	srand(l);
}
#endif

#ifdef WINDOWS
void gettimeofday(struct timeval* tv, void* null) {
    int tc = GetTickCount();
    tv->tv_sec = tc / 1000;
    tv->tv_usec = (tc % 1000) * 1000;
}
#endif

double ms_between(struct timeval *tv1, struct timeval *tv2) {
  return (tv2->tv_sec - tv1->tv_sec)*1000.0 + (tv2->tv_usec - tv1->tv_usec)*0.001;
}


double* compare_weights_vector = NULL;
INLINE int compare_weights(const void* v1, const void* v2) {
  int* i1 = (int*)v1;
  int* i2 = (int*)v2;
  double d1, d2;
  d1 = compare_weights_vector[ *i1 ];
  d2 = compare_weights_vector[ *i2 ];
  if (d1 < d2)
    return 1;
  else if (d1 > d2)
    return -1;
  else
    return 0;
}

INLINE int compare_weights_asc(const void* v1, const void* v2) {
  int* i1 = (int*)v1;
  int* i2 = (int*)v2;
  double d1, d2;
  d1 = compare_weights_vector[ *i1 ];
  d2 = compare_weights_vector[ *i2 ];
  if (d1 < d2)
    return -1;
  else if (d1 > d2)
    return 1;
  else
    return 0;
}

INLINE void permute_doubles(double* data, int* order, int N) {
  int i;
  double* temp = (double*) malloc(N * sizeof(double));
  for (i=0; i<N; i++) {
    temp[i] = data[order[i]];
  }
  for (i=0; i<N; i++) {
    data[i] = temp[i];
  }
  free(temp);
}

INLINE void permute_nodes(Node** data, int* order, int N) {
  int i;
  Node** temp = (Node**) malloc(N * sizeof(Node*));
  for (i=0; i<N; i++) {
    temp[i] = data[order[i]];
  }
  for (i=0; i<N; i++) {
    data[i] = temp[i];
  }
  free(temp);
}

INLINE void permute_ints(int* data, int* order, int N) {
  int i;
  int* temp = (int*) malloc(N * sizeof(int));
  for (i=0; i<N; i++) {
    temp[i] = data[order[i]];
  }
  for (i=0; i<N; i++) {
    data[i] = temp[i];
  }
  free(temp);
}

void setup_XYw(double* X, int NX, double* Y, int NY, int D, double* w, double x_low,
               double x_high, double y_low, double y_high, char* weights_mode) {
    int i;
    double x_range = x_high - x_low;
    double y_range = y_high - y_low;
    srand48((unsigned int) time(NULL));
//    srand48(1);
    for (i = 0; i < (NX*D); i++) {
        X[i] = x_low + x_range * drand48();
    }
    for (i = 0; i < (NY*D); i++) {
        Y[i] = y_low + y_range * drand48();
    }
    if (strcmp(weights_mode, "positive") == 0) {
        for (i = 0; i < NX; i++) {
            w[i] = drand48();
        }
    } else if (strcmp(weights_mode, "negative") == 0) {
        for (i = 0; i < NX; i++) {
            w[i] = -drand48();
        }
    } else if (strcmp(weights_mode, "mixed") == 0) {
        for (i = 0; i < NX; i++) {
            w[i] = -1 + 2 * drand48();
        }
    }
}

void setup_kernel(char* name) {
    if (strcmp(name, "gaussian") == 0) {
        kernel_init = &gauss_kernel_init;
        kernel_destroy = &gauss_kernel_destroy;
        kernel_influence = &gauss_influence;
        kernel_distance = &gauss_distance;
        kernel_increasing = &gauss_increasing;
    } else if (strcmp(name, "laplacian") == 0) {
        kernel_init = &laplace_kernel_init;
        kernel_destroy = &laplace_kernel_destroy;
        kernel_influence = &laplace_influence;
        kernel_distance = &laplace_distance;
        kernel_increasing = &laplace_increasing;
    } else if (strcmp(name, "lpq") == 0) {
        kernel_init = &lpq_kernel_init;
        kernel_destroy = &lpq_kernel_destroy;
        kernel_influence = &lpq_influence;
        kernel_distance = &lpq_distance;
        kernel_increasing = &lpq_increasing;
    } else if (strcmp(name, "sigmoid") == 0) {
        kernel_init = &sigmoid_kernel_init;
        kernel_destroy = &sigmoid_kernel_destroy;
        kernel_influence = &sigmoid_influence;
        kernel_distance = &sigmoid_distance;
        kernel_increasing = &sigmoid_increasing;
    } else if (strcmp(name, "polynomial") == 0) {
        kernel_init = &polynomial_kernel_init;
        kernel_destroy = &polynomial_kernel_destroy;
        kernel_influence = &polynomial_influence;
        kernel_distance = &polynomial_distance;
        kernel_increasing = &polynomial_increasing;
    } else if (strcmp(name, "dog") == 0) {
        kernel_init = &dog_kernel_init;
        kernel_destroy = &dog_kernel_destroy;
        kernel_influence = &dog_influence;
        kernel_distance = &dog_distance;
        kernel_increasing = &dog_increasing;
    } else if (strcmp(name, "gaussian_neg") == 0) {
        kernel_init = &gauss_neg_kernel_init;
        kernel_destroy = &gauss_neg_kernel_destroy;
        kernel_influence = &gauss_neg_influence;
        kernel_distance = &gauss_neg_distance;
        kernel_increasing = &gauss_neg_increasing;
    } else {
        return;
    }
}

int verification_check_sum(double* mginfls, double* naiveinfls, int NY, double eps_abs,
                        double eps_rel) {
    int ok = 1;
    int y;
    double avgerr = 0.0;
    double avgabserr = 0.0;
    double avgval = 0.0;
    double maxerr = 0.0;
    for (y = 0; y < NY; y++) {
        double diff = mginfls[y] - naiveinfls[y];
        double err = dabs(diff);
        double val = naiveinfls[y];
        avgerr += diff;
        avgabserr += err;
        avgval += val;
        if (err > eps_abs && err / dabs(val) > eps_rel) {
            printf("disagreement: y=%i, naive says %g, approx says %g (error %g, rel %g, eps_abs %g, eps_rel %g)\n",
                   y, naiveinfls[y], mginfls[y], err, err/dabs(val), eps_abs, eps_rel);
            ok = 0;
        }
        if (err > maxerr) {
            maxerr = err;
        }
    }
    avgval /= (double) NY;
    avgerr /= (double) NY;
    avgabserr /= (double) NY;
    if (!ok) {
        printf("Verification check FAILED!\n");
        printf("Average error: %g (target %g)\n", avgabserr, eps_abs);
        printf("Average systematic error: %g.\n", avgerr);
        printf("Average value: %g\n", avgval);
        printf("Maximum error: %g (target %g)\n", maxerr, eps_abs);
    }
//    } else {
//        printf("Verification check passed\n");
//        printf("Average error: %g (target %g)\n", avgabserr, eps_abs);
//        printf("Average systematic error: %g.\n", avgerr);
//        printf("Average value: %g\n", avgval);
//        printf("Maximum error: %g (target %g)\n", maxerr, eps_abs);
//    }
    return ok;
}

int verification_check_max(int* mginds, int* naiveinds, double* mginfls, double* naiveinfls, int NY) {
    int ok;
    int y;
    ok = 1;
    for (y = 0; y < NY; y++) {
        if (naiveinfls[y] != mginfls[y]) {
            printf("influence disagreement: y=%d, naive says %g, approx says %g, diff = %g.\n",
                   y, naiveinfls[y], mginfls[y], naiveinfls[y] - mginfls[y]);
            ok = 0;
        }
        if (naiveinds[y] != mginds[y]) {
            if (naiveinfls[y] == mginfls[y]) {
                // different points have same influence
                printf("different source particles have the same influence: %d\n", y);
                printf("naive says %d, approx says %d.\n", naiveinds[y], mginds[y]);
                printf("naive infl = %f, approx infl = %f\n", naiveinfls[y], mginfls[y]);
            } else {
                // something is wrong!
                printf("wrong indices: naive says %d, approx says %d.\n", naiveinds[y], mginds[y]);
            }
        }
        if (mginds[y] < 0) {
            printf("null indices!!\n");
        }
    }
    if (!ok) {
        printf("Verification check FAILED!\n");
    }
    return ok;
}
