/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */



#include "definitions.h"

#include "generic.h"

double max_influence_upper_bounds(Node* xnode, Node* ynode, tree_globals* globals);
double max_influence_lower_bounds(Node* xnode, Node* ynode, tree_globals* globals);
double max_influence_upper_bounds_p(Node* xnode, double* yp, tree_globals* globals);
double translate_influence(double infl, double maxweight, double weight);


#include "utils.h"
#include "generic_sum.h"

void generic_sum_helper(tree_globals* globals, double epsilon_relative, double epsilon_absolute, double* estimates, double* bounds, int* satisfied, int NY);

void sum_influence_bounds(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower);
void sib_mw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower);
void sib_nw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower);
void sib_pw(Node* xnode, Node* ynode, tree_globals* globals, double* upper, double* lower);

#include "generic_max.h"
#include "kernel.h"
#include "kernel_gaussian.h"
#include "kernel_laplacian.h"
#include "kernel_lpq.h"
#include "kernel_sigmoid.h"
#include "kernel_polynomial.h"
#include "kernel_dog.h"
#include "kernel_gaussian_neg.h"

#include "kdtree.h"
#include "dualtree.h"
#include "anchors.h"

void merge_anchors(double* X, int D, Node** anchors, int Nanchors);

#include "nodelist.h"
#include "intlist.h"
#include "doublelist.h"
#include "anchors.h"

void resort_points(Node* a);
void split_anchor_helper(nodelist* splittable, nodelist* anchors, double* X, int D, int Nmin, int Nanchors);
void split_anchor(Node* main, double* X, int D, int Nmin, int Nanchors, Node** result, int* Nresult);
void steal_points(double* X, int D, Node** nodes, int Ndone, Node* newone);
void remove_points(Node* node, intlist* stolenlist);

#include "kernel.h"
#include "anchor_max.h"
#include "boxtree_max.h"
#include "naive_max.h"
#include "anchor_sum.h"
#include "boxtree_sum.h"
#include "naive_sum.h"
#include "boxtree.h"

void place_in_boxes(Node* node, int N, double* X, int D, int d, int NB);

#include "distance_transform.h"
#include "gauss.h"
#include "pqueue.h"

#include "boxtree.c"
#include "wirth.c"
#include "pqueue.c"


#include "generic.c"
#include "generic_max.c"
#include "generic_sum.c"
#include "kernel_gaussian.c"
#include "kernel_laplacian.c"
#include "kernel_lpq.c"
#include "kernel_sigmoid.c"
#include "kernel_polynomial.c"
#include "kernel_dog.c"
#include "kernel_gaussian_neg.c"
#include "utils.c"

#include "nodelist.c"
#include "intlist.c"
#include "doublelist.c"
#include "anchors.c"
#include "anchor_max.c"

#include "boxtree_max.c"
#include "naive_max.c"
#include "anchor_sum.c"
#include "boxtree_sum.c"
#include "naive_sum.c"
#include "distance_transform.c"
#include "gauss.c"
#include "dualtree.c"

#include "kdtree.c"



void do_sum(double* X, double* Y, double* w, double* P, int NP, char* ker_type, int NX,
			int NY, int D, double eps_rel, double eps_abs, char* structure, double* infls,
			int influence_mode, int weights_mode) {
	struct timeval tv1;
	struct timeval tv2;
	double ms;
	/* printf("%s sum:\n", structure); */
    gettimeofday(&tv1, NULL);
    nbody_sum(X, Y, w, P, NP, ker_type, NX, NY, D, eps_rel, eps_abs, structure, infls, influence_mode, weights_mode);
    gettimeofday(&tv2, NULL);
    ms = ms_between(&tv1, &tv2);
    /* printf("%s sum took %f ms.\n", structure, ms); */
}

void do_max(double* X, double* Y, double* w, double* P, int NP, char* ker_type, int NX,
			int NY, int D, char* structure, int* inds, double* infls, int influence_mode, int weights_mode) {
	struct timeval tv1;
	struct timeval tv2;
	double ms;
	printf("%s max:\n", structure);
    gettimeofday(&tv1, NULL);
    nbody_max(X, Y, w, P, NP, ker_type, NX, NY, D, 25, structure, inds, infls, influence_mode, weights_mode);
    gettimeofday(&tv2, NULL);
    ms = ms_between(&tv1, &tv2);
    printf("%s max took %f ms.\n", structure, ms);
}

void resetData(int* mginds, double* mginfls, int NY) {
    int i;
    for (i = 0; i < NY; i++) {
        mginds[i] = -1;
        mginfls[i] = -INF;
    }
}


int covariance_matrix_vector_multiplication_fast_cmvps(double *X, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *v, double *result, int *TYPE, double *TOL_ABS, double *TOL_REL) {
    int NX = *n;
    int NY = NX;
    int D = *d;
    int np = 0;
	double params[4];
    char* ker_type = "gaussian";
    int influence_mode = MIXED_INFLUENCE;
    int weights_mode = MIXED_WEIGHTS;
    char* wm = "positive";
    double eps_rel = *TOL_REL;
    double eps_abs = *TOL_ABS;
    int ok = 1;

    int i;
    /* double* X; */
    /* double* Y; */
    double* w;
    w = (double*) malloc(NX * sizeof(double));
    for (i = 0; i < NX; i++) w[i] = v[i];

    double *mginfls, *naiveinfls;
    int *mginds, *naiveinds;

    /* X = (double*) malloc(NX * D * sizeof(double)); */
    /* Y = (double*) malloc(NY * D * sizeof(double)); */
  
    /* setup_XYw(X, NX, Y, NY, D, w, 0, 1, 0, 1, wm); */

    mginds = (int*) malloc(NY * sizeof(int));
    naiveinds = (int*) malloc(NY * sizeof(int));

    mginfls = (double*) malloc(NY * sizeof(double));
    naiveinfls = (double*) malloc(NY * sizeof(double));
    
    for (i = 0; i < NY; i++) {
        mginds[i] = -1;
        naiveinds[i] = -1;
        mginfls[i] = -INF;
        naiveinfls[i] = -INF;
    }
    

    /* printf("NX = %i.  NY = %i.  D = %i.\n", NX, NY, D); */
    
    /* printf("\nSUM\n\n"); */
    /* if (strcmp(ker_type, "gaussian") == 0) { */
        np = 1;
        params[0] = exp(- 0.5 * kpar[2]);
    /* }  */
	/* else if (strcmp(ker_type, "laplacian") == 0) { */
    /*     np = 1; */
    /*     params[0] = 0.1; */
    /* } else if (strcmp(ker_type, "lpq") == 0) { */
    /*     np = 3; */
    /*     params[0] = 25; params[1] = 2; params[2] = 1; */
    /* } else if (strcmp(ker_type, "sigmoid") == 0) { */
    /*     np = 4; */
    /*     params[0] = 2; params[1] = 1; params[2] = 1; params[3] = 4; */
    /* } else if (strcmp(ker_type, "polynomial") == 0) { */
    /*     np = 4; */
    /*     params[0] = 2; params[1] = 2; params[2] = 1; params[3] = 1; */
    /* } else if (strcmp(ker_type, "dog") == 0) { */
    /*     np = 1; */
    /*     params[0] = 3; */
    /* } else if (strcmp(ker_type, "gaussian_neg") == 0) { */
    /*     np = 1; */
    /*     params[0] = 2; */
    /* } */

		if(*TYPE == 1) {    
		  do_sum(X, X, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "naive", naiveinfls, influence_mode, weights_mode);

		  for(int i=0; i < *n; i++) {
			naiveinfls[i] *= exp_kpar[0];
			naiveinfls[i] += (*OMEGA + exp_kpar[1]) * v[i];
		  }
		  
		  for(int i=0; i < *n; i++) result[i] = naiveinfls[i];
		}
		
		if(*TYPE == 2) {
		  do_sum(X, X, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "anchors", mginfls, influence_mode, weights_mode);

		  for(int i=0; i < *n; i++) {
			mginfls[i] *= exp_kpar[0];
			mginfls[i] += (*OMEGA + exp_kpar[1]) * v[i];
		  }
		  
		  for(int i=0; i < *n; i++) result[i] = mginfls[i];
		}
    /* ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok; */
    /* resetData(mginds, mginfls, NY); */

		if(*TYPE == 3) {
		  do_sum(X, X, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "kdtree", mginfls, influence_mode, weights_mode);

		  for(int i=0; i < *n; i++) {
			mginfls[i] *= exp_kpar[0];
			mginfls[i] += (*OMEGA + exp_kpar[1]) * v[i];
		  }
		  
		  for(int i=0; i < *n; i++) result[i] = mginfls[i];
		}
    /* ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok; */
    /* resetData(mginds, mginfls, NY); */
    
    /* if (D <= 3 && strcmp(ker_type, "gaussian") == 0) { */
	/* 	do_sum(X, X, w, params, np, ker_type, NX, NY, D, eps_rel, eps_abs, "fgt", mginfls, influence_mode, weights_mode); */
    /*     ok = verification_check_sum(mginfls, naiveinfls, NY, eps_abs, eps_rel) && ok; */
    /*     resetData(mginds, mginfls, NY); */
    /* } */

        
    /* if (ok) { */
    /*     printf("\nAll passed!\n"); */
    /* } else { */
    /*     printf("\nSome FAILED!\n"); */
    /* } */

    free(w);
    free(mginds);
    free(mginfls);
    free(naiveinds);
    free(naiveinfls);

    return 0;
}
