#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

/* 
RBF kernel function

k(x, y | theta) = exp(theta[0]) * exp( - exp(theta[1]) * ||x - y ||^2 )

#### Input #### 
- (double *) pointer to input vector x
- (double *) pointer to input vector y
- (int) dimension d of the input vectors 
- (double *) pointer to vector of parameter theta

#### Output ####
- (double) RBF kernel between x and y for a given theta

*/

double calcKernel_gsl_by_data(double *DATA, gsl_matrix *X, int *n, int *d, unsigned int p, unsigned int q, double *kpar, double *exp_kpar, double *OMEGA) {
  double norm = 0;
  double diff;
  
  for (unsigned int i = 0; i < *d; i++) {
	diff = gsl_matrix_get(X, p, i) - gsl_matrix_get(X, q, i);
	norm += diff * diff;
  }
    
  return exp(-exp_kpar[2] * norm + kpar[0]);
}

double calcKernel_gsl_by_feat(double *DATA, gsl_matrix *X, int *n, int *d, unsigned int p, unsigned int q, double *kpar, double *exp_kpar, double *OMEGA) {
  double norm = 0;
  double diff;
  
  for (unsigned int i = 0; i < *d; i++) {
	diff = gsl_matrix_get(X, i, p) - gsl_matrix_get(X, i, q);
	norm += diff * diff;
  }
    
  return exp(-exp_kpar[2] * norm + kpar[0]);
}


double calcKernel_no_gsl_by_data(double *DATA, gsl_matrix *X, int *n, int *d, unsigned int p, unsigned int q, double *kpar, double *exp_kpar, double *OMEGA) {
  double norm = 0;
  double diff;

  for (unsigned int i = 0; i < *d * *n; i += *n) {
	diff = DATA[p + i] - DATA[q + i];
	norm += diff * diff;
  }
    
  return exp(-exp_kpar[2] * norm + kpar[0]);
}

double calcKernel_no_gsl_by_feat(double *DATA, gsl_matrix *X, int *n, int *d, unsigned int p, unsigned int q, double *kpar, double *exp_kpar, double *OMEGA) {
  double norm = 0;
  double diff;
  
  p *= *d;
  q *= *d;
  
  for (unsigned int i = 0; i < *d; i++) {
	diff = DATA[i + p] - DATA[i + q];
	norm += diff * diff;
  }
  
  return exp(-exp_kpar[2] * norm + kpar[0]);
}



void multiply_K_vector(double *DATA, gsl_matrix *matrixData, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *vector, int *type, gsl_vector* result) {

  double (*kernelFunction)(double*, gsl_matrix*, int*, int*, unsigned int, unsigned int, double*, double*, double*);

  if(*type == 1) kernelFunction = calcKernel_gsl_by_data;
  if(*type == 2) kernelFunction = calcKernel_gsl_by_feat;
  if(*type == 3) kernelFunction = calcKernel_no_gsl_by_data;
  if(*type == 4) kernelFunction = calcKernel_no_gsl_by_feat;

  double covariance;

  double res[*n];
  for(unsigned int i = 0; i < *n; i++) res[i] = exp_kpar[0] * vector[i];

  for(unsigned int i = 0; i < *n; i++) {
    for(unsigned int j = i + 1; j < *n; j++) {
      covariance = kernelFunction(DATA, matrixData, n, d, i, j, kpar, exp_kpar, OMEGA);
      
      res[i] += covariance * vector[j];
      res[j] += covariance * vector[i];
    }
  }

  for(unsigned int i = 0; i < *n; i++) res[i] += (exp_kpar[1] + *OMEGA) * vector[i];

  for(unsigned int i = 0; i < *n; i++) gsl_vector_set(result, i, res[i]);
}


int wrapper_multiply_K_vector(double *DATA, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double* b, int *type, double *result) {

  int i, j;
  
  gsl_vector* b_gsl = gsl_vector_alloc(*n);
  /* for(i = 0; i < *n; i++) gsl_vector_set(b_gsl, i, b[i]); */

  gsl_vector* result_gsl = gsl_vector_alloc(*n);

  gsl_matrix* DATA_gsl;

  // gsl_by_data
  if(*type == 1) {
	DATA_gsl = gsl_matrix_alloc(*n, *d);
	for(i = 0; i < *n; i++) {
	  for(j = 0; j < *d; j++) {
		gsl_matrix_set(DATA_gsl, i, j, DATA[j * (*n) + i]);
	  }
	}
  }

  // gsl_by_feat
  if(*type == 2) {
	DATA_gsl = gsl_matrix_alloc(*d, *n);
	for(i = 0; i < *d; i++) {
	  for(j = 0; j < *n; j++) {
		gsl_matrix_set(DATA_gsl, i, j, DATA[j * (*d) + i]);
	  }
	}
  }

  if(*type == 3) DATA_gsl = gsl_matrix_alloc(1, 1);
  if(*type == 4) DATA_gsl = gsl_matrix_alloc(1, 1);


  /* clock_t start, end; */
  /* double cpu_time_used; */
  
  /* start = clock(); */
  multiply_K_vector(DATA, DATA_gsl, n, d, kpar, exp_kpar, OMEGA, b, type, result_gsl);
  /* end = clock(); */
  /* cpu_time_used = ((double) (end - start)) / CLOCKS_PER_SEC * 1e3; */
  /* printf("multiplication took %lf ms.\n", cpu_time_used); */
  
  for(i = 0; i < *n; i++) result[i] = gsl_vector_get(result_gsl, i);

  gsl_vector_free(b_gsl);
  gsl_matrix_free(DATA_gsl);
  gsl_vector_free(result_gsl);
}
